# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from unittest import mock

import fixtures
from testtools.matchers import Contains, Equals, FileExists, Not
from xdg import BaseDirectory

import tests
from snapcraft import file_utils, internal, storeapi
from snapcraft.internal import review_tools
from snapcraft.storeapi.errors import (
    StoreDeltaApplicationError,
    StoreUpDownError,
    StoreUploadError,
)

from . import FakeStoreCommandsBaseTestCase


class UploadCommandBaseTestCase(FakeStoreCommandsBaseTestCase):
    def setUp(self):
        super().setUp()

        self.snap_file = os.path.join(
            os.path.dirname(tests.__file__), "data", "test-snap.snap"
        )

        self.fake_review_tools_run = fixtures.MockPatch(
            "snapcraft.internal.review_tools.run"
        )
        self.useFixture(self.fake_review_tools_run)

        self.fake_review_tools_is_available = fixtures.MockPatch(
            "snapcraft.internal.review_tools.is_available", return_value=False
        )
        self.useFixture(self.fake_review_tools_is_available)


class UploadCommandTestCase(UploadCommandBaseTestCase):
    def test_upload_without_snap_must_raise_exception(self):
        result = self.run_command(["upload"])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("Usage:"))

    def test_upload_a_snap(self):
        # Upload
        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("Revision 19 of 'basic' created."))
        self.fake_store_upload.mock.assert_called_once_with(
            snap_name="basic",
            snap_filename=self.snap_file,
            built_at=None,
            channels=None,
            delta_format=None,
            delta_hash=None,
            source_hash=None,
            target_hash=None,
        )

    def test_review_tools_not_available(self):
        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains(
                "Install the review-tools from the Snap Store for enhanced "
                "checks before uploading this snap"
            ),
        )
        self.fake_review_tools_run.mock.assert_not_called()

    def test_upload_a_snap_review_tools_run_success(self):
        self.fake_review_tools_is_available.mock.return_value = True

        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_review_tools_run.mock.assert_called_once_with(
            snap_filename=self.snap_file
        )

    def test_upload_a_snap_review_tools_run_fail(self):
        self.fake_review_tools_is_available.mock.return_value = True
        self.fake_review_tools_run.mock.side_effect = review_tools.errors.ReviewError(
            {
                "snap.v2_functional": {"error": {}, "warn": {}},
                "snap.v2_security": {
                    "error": {
                        "security-snap-v2:security_issue": {
                            "text": "(NEEDS REVIEW) security message."
                        }
                    },
                    "warn": {},
                },
                "snap.v2_lint": {"error": {}, "warn": {}},
            }
        )

        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains(
                "Review Tools did not fully pass for this snap.\n"
                "Specific measures might need to be taken on the Snap Store before "
                "this snap can be fully accepted.\n"
                "Security Issues:\n"
                "- (NEEDS REVIEW) security message"
            ),
        )
        self.fake_review_tools_run.mock.assert_called_once_with(
            snap_filename=self.snap_file
        )

    def test_upload_with_started_at(self):
        snap_file = os.path.join(
            os.path.dirname(tests.__file__), "data", "test-snap-with-started-at.snap"
        )

        # Upload
        result = self.run_command(["upload", snap_file])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("Revision 19 of 'basic' created."))
        self.fake_store_upload.mock.assert_called_once_with(
            snap_name="basic",
            snap_filename=snap_file,
            built_at="2019-05-07T19:25:53.939041Z",
            channels=None,
            delta_format=None,
            delta_hash=None,
            source_hash=None,
            target_hash=None,
        )

    def test_upload_without_login_must_ask(self):
        self.fake_store_upload_precheck.mock.side_effect = [
            storeapi.http_clients.errors.InvalidCredentialsError("error"),
            None,
        ]

        result = self.run_command(
            ["upload", self.snap_file], input="\n\n\n\nuser@example.com\nsecret\n"
        )

        self.assertThat(
            result.output, Contains("You are required to login before continuing.")
        )

    def test_upload_nonexisting_snap_must_raise_exception(self):
        result = self.run_command(["upload", "test-unexisting-snap"])

        self.assertThat(result.exit_code, Equals(2))

    def test_upload_invalid_snap_must_raise_exception(self):
        snap_path = os.path.join(
            os.path.dirname(tests.__file__), "data", "invalid.snap"
        )

        raised = self.assertRaises(
            internal.errors.SnapDataExtractionError,
            self.run_command,
            ["upload", snap_path],
        )

        self.assertThat(str(raised), Contains("Cannot read data from snap"))

    def test_upload_unregistered_snap_must_ask(self):
        class MockResponse:
            status_code = 404

            def json(self):
                return dict(
                    error_list=[
                        {
                            "code": "resource-not-found",
                            "message": "Snap not found for name=basic",
                        }
                    ]
                )

        self.fake_store_upload_precheck.mock.side_effect = [
            StoreUploadError("basic", MockResponse()),
            None,
        ]

        result = self.run_command(["upload", self.snap_file], input="y\n")

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains("You are required to register this snap before continuing. "),
        )
        self.fake_store_register.mock.assert_called_once_with(
            "basic", is_private=False, series="16", store_id=None
        )

    def test_upload_unregistered_snap_must_raise_exception_if_not_registering(self):
        class MockResponse:
            status_code = 404

            def json(self):
                return dict(
                    error_list=[
                        {
                            "code": "resource-not-found",
                            "message": "Snap not found for name=basic",
                        }
                    ]
                )

        self.fake_store_upload_precheck.mock.side_effect = [
            StoreUploadError("basic", MockResponse()),
            None,
        ]

        raised = self.assertRaises(
            storeapi.errors.StoreUploadError,
            self.run_command,
            ["upload", self.snap_file],
        )

        self.assertThat(
            str(raised),
            Contains("This snap is not registered. Register the snap and try again."),
        )
        self.fake_store_register.mock.assert_not_called()

    def test_upload_with_updown_error(self):
        # We really don't know of a reason why this would fail
        # aside from a 5xx style error on the server.
        class MockResponse:
            text = "stub error"
            reason = "stub reason"

        self.fake_store_upload.mock.side_effect = StoreUpDownError(MockResponse())

        self.assertRaises(
            storeapi.errors.StoreUpDownError,
            self.run_command,
            ["upload", self.snap_file],
        )

    def test_upload_raises_deprecation_warning(self):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        # Push
        result = self.run_command(["push", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("Revision 19 of 'basic' created."))
        self.assertThat(
            fake_logger.output,
            Contains(
                "DEPRECATED: The 'push' set of commands have been replaced with 'upload'."
            ),
        )
        self.fake_store_upload.mock.assert_called_once_with(
            snap_name="basic",
            snap_filename=self.snap_file,
            built_at=None,
            channels=None,
            delta_format=None,
            delta_hash=None,
            source_hash=None,
            target_hash=None,
        )

    def test_upload_and_release_a_snap(self):
        self.useFixture
        # Upload
        result = self.run_command(["upload", self.snap_file, "--release", "beta"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("Revision 19 of 'basic' created"))
        self.fake_store_upload.mock.assert_called_once_with(
            snap_name="basic",
            snap_filename=self.snap_file,
            built_at=None,
            channels=["beta"],
            delta_format=None,
            delta_hash=None,
            source_hash=None,
            target_hash=None,
        )

    def test_upload_and_release_a_snap_to_N_channels(self):
        # Upload
        result = self.run_command(
            ["upload", self.snap_file, "--release", "edge,beta,candidate"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("Revision 19 of 'basic' created"))
        self.fake_store_upload.mock.assert_called_once_with(
            snap_name="basic",
            snap_filename=self.snap_file,
            built_at=None,
            channels=["edge", "beta", "candidate"],
            delta_format=None,
            delta_hash=None,
            source_hash=None,
            target_hash=None,
        )

    def test_upload_displays_humanized_message(self):
        result = self.run_command(
            ["upload", self.snap_file, "--release", "edge,beta,candidate"]
        )

        self.assertThat(
            result.output,
            Contains(
                "After uploading, the resulting snap revision will be released to "
                "'beta', 'candidate', and 'edge' when it passes the Snap Store review."
            ),
        )


class UploadCommandDeltasTestCase(UploadCommandBaseTestCase):
    def setUp(self):
        super().setUp()

        self.latest_snap_revision = 8
        self.new_snap_revision = self.latest_snap_revision + 1

        self.mock_tracker.track.return_value = {
            "code": "ready_to_release",
            "processed": True,
            "can_release": True,
            "url": "/fake/url",
            "revision": self.new_snap_revision,
        }

    def test_upload_revision_cached_with_experimental_deltas(self):
        # Upload
        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))
        snap_cache = os.path.join(
            BaseDirectory.xdg_cache_home,
            "snapcraft",
            "projects",
            "basic",
            "snap_hashes",
            "amd64",
        )
        cached_snap = os.path.join(
            snap_cache, file_utils.calculate_sha3_384(self.snap_file)
        )

        self.assertThat(cached_snap, FileExists())

    def test_upload_revision_uses_available_delta(self):
        # Upload
        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))

        # Upload again
        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))
        _, kwargs = self.fake_store_upload.mock.call_args
        self.assertThat(kwargs.get("delta_format"), Equals("xdelta3"))

    def test_upload_with_delta_generation_failure_falls_back(self):
        # Upload and ensure fallback is called
        with mock.patch(
            "snapcraft._store._upload_delta",
            side_effect=StoreDeltaApplicationError("error"),
        ):
            result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_store_upload.mock.assert_called_once_with(
            snap_name="basic",
            snap_filename=self.snap_file,
            built_at=None,
            channels=None,
            delta_format=None,
            delta_hash=None,
            source_hash=None,
            target_hash=None,
        )

    def test_upload_with_delta_upload_failure_falls_back(self):
        # Upload
        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))

        result = {
            "code": "processing_upload_delta_error",
            "errors": [{"message": "Delta service failed to apply delta within 60s"}],
        }
        self.mock_tracker.raise_for_code.side_effect = [
            storeapi.errors.StoreReviewError(result=result),
            None,
        ]

        # Upload and ensure fallback is called
        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_store_upload.mock.assert_has_calls(
            [
                mock.call(
                    snap_name="basic",
                    snap_filename=mock.ANY,
                    built_at=None,
                    channels=None,
                    delta_format="xdelta3",
                    delta_hash=mock.ANY,
                    source_hash=mock.ANY,
                    target_hash=mock.ANY,
                ),
                mock.call(
                    snap_name="basic",
                    snap_filename=self.snap_file,
                    built_at=None,
                    channels=None,
                    delta_format=None,
                    delta_hash=None,
                    source_hash=None,
                    target_hash=None,
                ),
            ]
        )

    def test_upload_with_disabled_delta_falls_back(self):
        # Upload
        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))

        class _FakeResponse:
            status_code = 501
            reason = "disabled"

            def json(self):
                return {
                    "error_list": [
                        {
                            "code": "feature-disabled",
                            "message": "The delta upload support is currently disabled.",
                        }
                    ]
                }

        self.fake_store_upload.mock.side_effect = [
            storeapi.http_clients.errors.StoreServerError(_FakeResponse()),
            self.mock_tracker,
        ]

        # Upload and ensure fallback is called
        with mock.patch("snapcraft.storeapi._status_tracker.StatusTracker"):
            result = self.run_command(["upload", self.snap_file])
        self.assertThat(result.exit_code, Equals(0))
        self.fake_store_upload.mock.assert_has_calls(
            [
                mock.call(
                    snap_name="basic",
                    snap_filename=mock.ANY,
                    built_at=None,
                    channels=None,
                    delta_format="xdelta3",
                    delta_hash=mock.ANY,
                    source_hash=mock.ANY,
                    target_hash=mock.ANY,
                ),
                mock.call(
                    snap_name="basic",
                    snap_filename=self.snap_file,
                    built_at=None,
                    channels=None,
                    delta_format=None,
                    delta_hash=None,
                    source_hash=None,
                    target_hash=None,
                ),
            ]
        )


class UploadCommandDeltasWithPruneTestCase(UploadCommandBaseTestCase):
    def run_test(self, cached_snaps):
        snap_revision = 19

        self.mock_tracker.track.return_value = {
            "code": "ready_to_release",
            "processed": True,
            "can_release": True,
            "url": "/fake/url",
            "revision": snap_revision,
        }

        deb_arch = "amd64"

        snap_cache = os.path.join(
            BaseDirectory.xdg_cache_home,
            "snapcraft",
            "projects",
            "basic",
            "snap_hashes",
            deb_arch,
        )
        os.makedirs(snap_cache)

        for cached_snap in cached_snaps:
            cached_snap = cached_snap.format(deb_arch)
            open(os.path.join(snap_cache, cached_snap), "a").close()

        # Upload
        result = self.run_command(["upload", self.snap_file])

        self.assertThat(result.exit_code, Equals(0))

        real_cached_snap = os.path.join(
            snap_cache, file_utils.calculate_sha3_384(self.snap_file)
        )

        self.assertThat(os.path.join(snap_cache, real_cached_snap), FileExists())

        for snap in cached_snaps:
            snap = snap.format(deb_arch)
            self.assertThat(os.path.join(snap_cache, snap), Not(FileExists()))
        self.assertThat(len(os.listdir(snap_cache)), Equals(1))

    def test_delete_other_cache_files_with_valid_name(self):
        self.run_test(
            ["a-cached-snap_0.3_{}_8.snap", "another-cached-snap_1.0_fakearch_6.snap"]
        )

    def test_delete_other_cache_file_with_invalid_name(self):
        self.run_test(
            [
                "a-cached-snap_0.3_{}.snap",
                "cached-snap-without-revision_1.0_fakearch.snap",
                "another-cached-snap-without-version_fakearch.snap",
            ]
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

import pytest
from testtools.matchers import Equals, HasLength

from snapcraft.internal import errors, meta
from snapcraft.plugins.v1 import waf
from snapcraft.project import Project
from tests import unit

from . import PluginsV1BaseTestCase


class WafPluginBaseTest(PluginsV1BaseTestCase):
    """Plugin to provide snapcraft support for the waf build system"""

    def setUp(self):
        super().setUp()

        class Options:
            """Internal Options Class matching the Waf plugin"""

            configflags = []

        self.options = Options()


class WafPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        """Test validity of the Waf Plugin schema"""
        schema = waf.WafPlugin.schema()

        # Verify the presence of all properties
        properties = schema["properties"]
        self.assertTrue(
            "configflags" in properties,
            'Expected "configflags" to be included in properties',
        )

        # Check configflags property
        configflags = properties["configflags"]
        for item in ["type", "minitems", "uniqueItems", "items", "default"]:
            self.assertTrue(
                item in configflags,
                'Expected "{}" to be included in "configflags"'.format(item),
            )

        configflags_type = configflags["type"]
        self.assertThat(
            configflags_type,
            Equals("array"),
            'Expected "configflags" "type" to be "array", but it '
            'was "{}"'.format(configflags_type),
        )

        configflags_minitems = configflags["minitems"]
        self.assertThat(
            configflags_minitems,
            Equals(1),
            'Expected "configflags" "minitems" to be 1, but '
            "it was {}".format(configflags_minitems),
        )

        self.assertTrue(configflags["uniqueItems"])

        configflags_default = configflags["default"]
        self.assertThat(
            configflags_default,
            Equals([]),
            'Expected "configflags" "default" to be [], but '
            "it was {}".format(configflags_default),
        )

        configflags_items = configflags["items"]
        self.assertTrue(
            "type" in configflags_items,
            'Expected "type" to be included in "configflags" ' '"items"',
        )

        configflags_items_type = configflags_items["type"]
        self.assertThat(
            configflags_items_type,
            Equals("string"),
            'Expected "configflags" "items" "type" to be '
            '"string", but it was "{}"'.format(configflags_items_type),
        )

    def test_get_build_properties(self):
        expected_build_properties = ["configflags"]
        resulting_build_properties = waf.WafPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)


class WafPluginTest(WafPluginBaseTest):
    def waf_build(self):
        """Helper to call a full build"""
        plugin = waf.WafPlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)

        # Create fake waf
        open(os.path.join(plugin.sourcedir, "waf"), "w").close()

        plugin.build()

        return plugin

    @mock.patch.object(waf.WafPlugin, "run")
    def test_build_with_destdir(self, run_mock):
        """Test building via waf and check for known calls and destdir"""
        plugin = self.waf_build()

        self.assertThat(run_mock.call_count, Equals(4))
        run_mock.assert_has_calls(
            [
                mock.call(["./waf", "distclean"]),
                mock.call(["./waf", "configure"]),
                mock.call(["./waf", "build"]),
                mock.call(
                    ["./waf", "install", "--destdir={}".format(plugin.installdir)]
                ),
            ]
        )


class WafPluginUnsupportedBase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            configflags = []

        self.options = Options()

    def test_unsupported_base(self):
        self.assertRaises(
            errors.PluginBaseError,
            waf.WafPlugin,
            "test-part",
            self.options,
            self.project,
        )


@pytest.mark.parametrize("deb_arch", ["armhf", "arm64", "i386", "ppc64el"])
def test_cross_compile(monkeypatch, deb_arch):
    monkeypatch.setattr(Project, "is_cross_compiling", True)

    class Options:
        configflags = []

    project = Project(target_deb_arch=deb_arch)
    project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

    plugin = waf.WafPlugin("test-part", Options(), project)
    plugin.enable_cross_compilation()

    env = plugin.env(plugin.sourcedir)
    assert f"CC={project.arch_triplet}-gcc" in env
    assert f"CXX={project.arch_triplet}-g++" in env



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import snapcraft


class TestPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {"test-property": {"type": "string"}},
        }

    @classmethod
    def get_pull_properties(cls):
        return ["test-property"]

    @classmethod
    def get_build_properties(cls):
        return ["test-property"]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class PluginImpl(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {"foo": {"type": "string"}},
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return {"gcc"}

    def get_build_environment(self) -> Dict[str, str]:
        return dict()

    def get_build_commands(self) -> List[str]:
        return [
            "mkdir -p ${SNAPCRAFT_PART_INSTALL}/bin",
            "gcc hello.c -o ${SNAPCRAFT_PART_INSTALL}/bin/hello",
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import Contains, DirExists, Not

from tests import integration


class CleanDependentsTestCase(integration.TestCase):
    def setUp(self):
        super().setUp()

        self.copy_project_to_cwd("dependencies")
        self.run_snapcraft("prime")

        # Need to use the state directory here instead of partdir due to
        # bug #1567054.
        self.state_dirs = {
            "p1": os.path.join(self.parts_dir, "p1", "state"),
            "p2": os.path.join(self.parts_dir, "p2", "state"),
            "p3": os.path.join(self.parts_dir, "p3", "state"),
            "p4": os.path.join(self.parts_dir, "p4", "state"),
        }

    def assert_clean(self, parts, common=False):
        for part in parts:
            self.expectThat(
                self.state_dirs[part],
                Not(DirExists()),
                "Expected state directory for {!r} to be cleaned".format(part),
            )

        if common:
            self.expectThat(
                self.parts_dir,
                Not(DirExists()),
                "Expected parts/ directory to be cleaned",
            )
            self.expectThat(
                self.stage_dir,
                Not(DirExists()),
                "Expected stage/ directory to be cleaned",
            )
            self.expectThat(
                self.prime_dir,
                Not(DirExists()),
                "Expected prime/ directory to be cleaned",
            )

    def assert_not_clean(self, parts, common=False):
        for part in parts:
            self.expectThat(
                self.state_dirs[part],
                DirExists(),
                "Expected state directory for {!r} to not be cleaned".format(part),
            )

        if common:
            self.expectThat(
                self.parts_dir,
                DirExists(),
                "Expected parts/ directory to not be cleaned",
            )
            self.expectThat(
                self.stage_dir,
                DirExists(),
                "Expected stage/ directory to not be cleaned",
            )
            self.expectThat(
                self.prime_dir,
                DirExists(),
                "Expected prime/ directory to not be cleaned",
            )

    def test_clean_nested_dependent(self):
        # Test that p3 (which has dependencies but no dependents) cleans with
        # no extra parameters.
        self.run_snapcraft(["clean", "p3"])
        self.assert_clean(["p3"])
        self.assert_not_clean(["p1", "p2", "p4"], True)

        # Now run prime again
        self.run_snapcraft("prime")
        self.assert_not_clean(["p1", "p2", "p3", "p4"], True)

    def test_clean_dependent(self):
        # Test that p2 (which has both dependencies and dependents) cleans with
        # its dependents (p3 and p4) also specified.
        self.run_snapcraft(["clean", "p2", "p3", "p4"])
        self.assert_clean(["p2", "p3", "p4"])
        self.assert_not_clean(["p1"], True)

        # Now run prime again
        self.run_snapcraft("prime")
        self.assert_not_clean(["p1", "p2", "p3", "p4"], True)

    def test_clean_main(self):
        # Test that p1 (which has no dependencies but dependents) cleans with
        # its dependents (p2 and, as an extension, p3 and p4) also specified.
        self.run_snapcraft(["clean", "p1", "p2", "p3", "p4"])
        self.assert_clean(["p1", "p2", "p3", "p4"], True)

        # Now run prime again
        self.run_snapcraft("prime")
        self.assert_not_clean(["p1", "p2", "p3", "p4"], True)

    def test_clean_dependent_without_nested_dependents(self):
        # Test that cleaning p2 marks dependents as dirty, but doesn't clean
        # them
        output = self.run_snapcraft(["clean", "p2"])
        self.assertThat(
            output,
            Contains(
                "Cleaned 'p2', which makes the following parts out of date: 'p3' "
                "and 'p4'"
            ),
        )
        self.assert_not_clean(["p1", "p3", "p4"])
        self.assert_clean(["p2"])

    def test_clean_dependent_without_nested_dependent(self):
        # Test that cleaning p2 marks dependents as dirty, but doesn't clean
        # them if they're not specified
        output = self.run_snapcraft(["clean", "p2", "p3"])
        self.assertThat(
            output,
            Contains("Cleaned 'p2', which makes the following part out of date: 'p4'"),
        )
        self.assert_not_clean(["p1", "p4"])
        self.assert_clean(["p2", "p3"])

    def test_clean_main_without_any_dependent(self):
        # Test that p1 (which has no dependencies but dependents)
        # can be cleaned, but marks everything else as dirty
        output = self.run_snapcraft(["clean", "p1"])
        self.assertThat(
            output,
            Contains(
                "Cleaned 'p1', which makes the following parts out of date: 'p2', "
                "'p3', and 'p4'"
            ),
        )
        self.assert_clean(["p1"])
        self.assert_not_clean(["p2", "p3", "p4"])

    def test_clean_main_without_dependent(self):
        # Test that p1 and its nested dependencies can be cleaned while p2 is
        # marked as dirty
        output = self.run_snapcraft(["clean", "p1", "p3", "p4"])
        self.assertThat(
            output,
            Contains("Cleaned 'p1', which makes the following part out of date: 'p2'"),
        )
        self.assert_clean(["p1", "p3", "p4"])
        self.assert_not_clean(["p2"])

    def test_clean_main_without_nested_dependent(self):
        # Test that p1 and p2 clean and mark the nested dependents as dirty,
        # but don't clean them
        output = self.run_snapcraft(["clean", "p1", "p2"])
        self.assertThat(
            output,
            Contains(
                "Cleaned 'p2', which makes the following parts out of date: 'p3' "
                "and 'p4'"
            ),
        )
        self.assert_clean(["p1", "p2"])
        self.assert_not_clean(["p3", "p4"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, Optional

from snapcraft.internal.errors import SnapcraftException


class ReviewToolMissing(SnapcraftException):
    def get_brief(self) -> str:
        return "The Review Tools are not installed on this system."

    def get_resolution(self) -> str:
        return "Review Tools can be installed with:\nsnap install review-tools"


class ReviewError(SnapcraftException):
    def __init__(self, review_json: Dict[str, Any]) -> None:
        self._review_json = review_json

    def get_brief(self) -> str:
        return "Review Tools did not fully pass for this snap."

    def get_resolution(self) -> str:
        return (
            "Specific measures might need to be taken on the Snap Store before "
            "this snap can be fully accepted."
        )

    def _get_issues(self, issue_type: str) -> Dict[str, Any]:
        issues: Dict[str, Any] = dict()
        for severity in ("error", "warn"):
            issues.update(self._review_json[issue_type].get(severity))

        return issues

    def get_details(self) -> Optional[str]:
        all_issues = {
            "Linting Issues": self._get_issues("snap.v2_lint"),
            "Functional Issues": self._get_issues("snap.v2_functional"),
            "Security Issues": self._get_issues("snap.v2_security"),
        }

        # Reduce to actual issues.
        existing_issues = {k: v for k, v in all_issues.items() if v is not None}

        # Return early if there are no details.
        if not existing_issues:
            return None

        details = ""
        for issue_title, issues in sorted(existing_issues.items()):
            if not issues:
                continue
            details += f"{issue_title}:\n"
            for issue in issues.values():
                if "link" in issue:
                    details += "- {text} (Refer to {link})\n".format(**issue)
                else:
                    details += "- {text}\n".format(**issue)
            else:
                details += "\n"

        return details.strip()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools.matchers import Equals, HasLength

from snapcraft.storeapi import channels, errors, status
from tests import unit


class TestSnapStatusChannelDetails:
    scenarios = (
        (
            "specific",
            dict(
                payload={
                    "version": "3.4",
                    "channel": "stable",
                    "info": "specific",
                    "revision": 1,
                },
                expected_version="3.4",
                expected_channel="stable",
                expected_info="specific",
                expected_revision=1,
            ),
        ),
        (
            "tracking",
            dict(
                payload={"channel": "candidate", "info": "tracking"},
                expected_version=None,
                expected_channel="candidate",
                expected_info="tracking",
                expected_revision=None,
            ),
        ),
        (
            "specific",
            dict(
                payload={"channel": "stable", "info": "none"},
                expected_version=None,
                expected_channel="stable",
                expected_info="none",
                expected_revision=None,
            ),
        ),
    )

    def test_properties(
        self,
        payload,
        expected_version,
        expected_channel,
        expected_info,
        expected_revision,
    ):
        s = status.SnapStatusChannelDetails(
            snap_name="test-snap", arch="arm64", payload=payload
        )

        assert s.arch == "arm64"
        assert s.version == expected_version
        assert s.channel == expected_channel
        assert s.info == expected_info
        assert repr(s) == "<SnapStatusChannelDetails: test-snap on arm64>"

        assert list(s) == ["arm64", expected_revision, expected_version]


class SnapStatusTrackTest(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.status = status.SnapStatusTrackDetails(
            snap_name="test-snap",
            track="latest",
            payload={
                "ppc64el": [
                    {
                        "version": "3.4",
                        "channel": "stable",
                        "info": "specific",
                        "revision": 2833,
                    },
                    {"channel": "candidate", "info": "tracking"},
                    {"channel": "beta", "info": "tracking"},
                    {
                        "version": "3.4+git11.g8d7f165",
                        "channel": "edge",
                        "info": "specific",
                        "revision": 2895,
                    },
                ]
            },
        )

    def test_properties(self):
        self.assertThat(self.status.track, Equals("latest"))

    def test_repr(self):
        self.assertThat(
            repr(self.status), Equals("<SnapStatusTrackDetails: test-snap on latest>")
        )

    def test_get_arches(self):
        self.assertThat(self.status.get_arches(), Equals(["ppc64el"]))

    def test_get_channel(self):
        c = self.status.get_channel(risk="candidate", arch="ppc64el")
        self.assertThat(c.arch, Equals("ppc64el"))
        self.assertThat(c.version, Equals(None))
        self.assertThat(c.channel, Equals("candidate"))
        self.assertThat(c.revision, Equals(None))

    def test_get_missing_arch(self):
        self.assertRaises(
            errors.ChannelNotAvailableOnArchError,
            self.status.get_channel,
            risk="candidate",
            arch="amd64",
        )

    def test_get_missing_branch(self):
        self.assertRaises(
            errors.ChannelNotAvailableOnArchError,
            self.status.get_channel,
            risk="candidate",
            branch="patch-1",
            arch="ppcel64",
        )


class SnapStatusTest(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.status = status.SnapStatus(
            snap_name="test-snap",
            payload={
                "channel_map_tree": {
                    "latest": {
                        "16": {
                            "amd64": [
                                {"channel": "stable", "info": "none"},
                                {"channel": "candidate", "info": "none"},
                                {"channel": "beta", "info": "none"},
                                {
                                    "channel": "edge",
                                    "info": "specific",
                                    "revision": 1,
                                    "version": "0.1",
                                },
                            ]
                        }
                    }
                }
            },
        )

    def test_repr(self):
        self.assertThat(repr(self.status), Equals("<SnapStatus: test-snap>"))

    def test_get_tracks(self):
        self.assertThat(self.status.get_tracks(), Equals(["latest"]))

    def test_get_track(self):
        track_details = self.status.get_track("latest")
        self.assertThat(track_details.track, Equals("latest"))
        self.assertThat(
            track_details.get_channel(risk="stable", arch="amd64").channel,
            Equals("stable"),
        )

    def test_get_channel_set(self):
        channel_set = self.status.get_channel_set(channels.Channel("edge"))
        self.assertThat(channel_set, HasLength(1))

    def test_get_tracking_channel_set(self):
        self.assertRaises(
            errors.InvalidChannelSet,
            self.status.get_channel_set,
            channels.Channel("candidate"),
        )

    def test_get_none_channel_set(self):
        self.assertRaises(
            errors.InvalidChannelSet,
            self.status.get_channel_set,
            channels.Channel("stable"),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import pathlib
from typing import Any, Dict, List, Type

import tinydb
import yaml

import snapcraft

from . import errors, migration

logger = logging.getLogger(__name__)


class _YAMLStorage(tinydb.Storage):
    """Provide YAML-backed storage for TinyDB."""

    def __init__(self, path: str):
        self.path = pathlib.Path(path)
        logger.debug(f"_YAMLStorage init {self.path}")

    def read(self) -> Dict[str, Any]:
        """Read database from file."""
        logger.debug(f"_YAMLStorage read: {self.path}")
        try:
            with self.path.open() as fd:
                db_data = yaml.safe_load(fd)
        except FileNotFoundError:
            return dict()

        if not isinstance(db_data, dict) or any(
            [not isinstance(k, str) for k in db_data.keys()]
        ):
            raise RuntimeError(
                f"Invalid datastore contents for {str(self.path)}: {db_data!r}"
            )

        return db_data

    def write(self, data) -> None:
        """Write database (data) to file."""
        logger.debug(f"_YAMLStorage write: {self.path} data={data!r}")
        self.path.write_text(yaml.dump(data))

    def close(self):
        """Nothing to do since we do not keep <path> open."""
        logger.debug(f"_YAMLStorage close: {self.path}")


class _YAMLStorageReadOnly(_YAMLStorage):
    def write(self, data) -> None:
        """Ignore any writes in read-only mode."""


class Datastore:
    """Datastore class, providing context manager for TinyDB.

    Manages migrations and storage requirements.  If migrations
    do not indicate support for current datastore version,
    SnapcraftDatastoreVersionUnsupported will be raised. In that
    event, some basic fallback mode can be utilized by re-opening
    datastore in read-only mode."""

    def __init__(
        self,
        *,
        path: pathlib.Path,
        migrations: List[Type[migration.Migration]],
        read_only: bool = False,
        snapcraft_version: str = snapcraft.__version__,
    ) -> None:
        self.path = path
        self._snapcraft_version = snapcraft_version

        if read_only:
            storage_class: Type[_YAMLStorage] = _YAMLStorageReadOnly
        else:
            storage_class = _YAMLStorage

        self.db = tinydb.TinyDB(str(path), storage=storage_class)

        logger.debug(f"Datastore init: {self.path} read_only={read_only}")

        # Force the datastore to be read by making a query, otherwise it is
        # only read on the first access.
        _ = self.db.tables()

        # Nothing left to do if opening in read-only mode.
        if read_only:
            return

        current_version: int = 0
        supported_version: int = 0

        for migration_class in migrations:
            current_version = migration_class(
                db=self.db, snapcraft_version=self._snapcraft_version
            ).apply()
            supported_version = migration_class.SCHEMA_VERSION

        if current_version > supported_version:
            raise errors.SnapcraftDatastoreVersionUnsupported(
                path=self.path,
                current_version=current_version,
                supported_version=supported_version,
            )

    def __enter__(self) -> tinydb.TinyDB:
        return self.db

    def __exit__(self, exc_value, exc_type, exc_traceback) -> None:
        self.close()

    def close(self) -> None:
        """Close database."""
        self.db.close()
        logger.debug(f"Datastore close: {self.path}")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from snapcraft.internal.project_loader._extensions.ros2_foxy import (
    ExtensionImpl as Ros2FoxyExtension,
)


@pytest.fixture(params=[Ros2FoxyExtension])
def extension_class(request):
    return request.param


def test_extension(extension_class):
    ros2_extension = extension_class(
        extension_name="ros2-foxy", yaml_data=dict(base="core20")
    )

    assert ros2_extension.root_snippet == {
        "package-repositories": [
            {
                "components": ["main"],
                "formats": ["deb"],
                "key-id": "C1CF6E31E6BADE8868B172B4F42ED6FBAB17C654",
                "key-server": "keyserver.ubuntu.com",
                "suites": ["focal"],
                "type": "apt",
                "url": "http://repo.ros2.org/ubuntu/main",
            }
        ]
    }

    assert ros2_extension.app_snippet == {
        "command-chain": ["snap/command-chain/ros2-launch"],
        "environment": {
            "PYTHONPATH": "$SNAP/opt/ros/foxy/lib/python3.8/site-packages:$SNAP/usr/lib/python3/dist-packages:${PYTHONPATH}",
            "ROS_DISTRO": "foxy",
        },
    }

    assert ros2_extension.part_snippet == {
        "build-environment": [{"ROS_DISTRO": "foxy"}]
    }

    assert ros2_extension.parts == {
        "ros2-foxy-extension": {
            "build-packages": ["ros-foxy-ros-core"],
            "override-build": "install -D -m 0755 launch "
            "${SNAPCRAFT_PART_INSTALL}/snap/command-chain/ros2-launch",
            "plugin": "nil",
            "source": "$SNAPCRAFT_EXTENSIONS_DIR/ros2",
        }
    }


def test_supported_bases(extension_class):
    assert extension_class.get_supported_bases() == ("core20",)


def test_supported_confinement(extension_class):
    extension_class.get_supported_confinement() == ("strict", "devmode")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import shutil
import subprocess
from unittest import mock

import fixtures
from testtools.matchers import Contains, Equals, FileExists, Not

import tests
from snapcraft import internal, storeapi

from . import CommandBaseTestCase


class SnapTest(fixtures.TempDir):
    """Test fixture for copying testing snaps in a temporary directory.

    So they can be manipulated in isolation and all the testing mess
    gets cleaned up automatically.
    """

    data_dir = os.path.join(os.path.dirname(tests.__file__), "data")

    def __init__(self, test_snap_name):
        super(SnapTest, self).__init__()
        self.test_snap_name = test_snap_name

    def _setUp(self):
        super(SnapTest, self)._setUp()
        test_snap_path = os.path.join(self.data_dir, self.test_snap_name)
        self.snap_path = os.path.join(self.path, self.test_snap_name)
        shutil.copyfile(test_snap_path, self.snap_path)


class SignBuildTestCase(CommandBaseTestCase):
    def setUp(self):
        super().setUp()
        self.snap_test = SnapTest("test-snap.snap")
        self.useFixture(self.snap_test)

    def test_sign_build_nonexisting_snap(self):
        result = self.run_command(["sign-build", "nonexisting.snap"])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(
            result.output,
            Contains(
                "Error: Invalid value for '<snap-file>': File 'nonexisting.snap' does not exist.\n"
            ),
        )

    def test_sign_build_invalid_snap(self):
        snap_path = os.path.join(
            os.path.dirname(tests.__file__), "data", "invalid.snap"
        )

        raised = self.assertRaises(
            internal.errors.SnapDataExtractionError,
            self.run_command,
            ["sign-build", snap_path],
        )

        self.assertThat(str(raised), Contains("Cannot read data from snap"))

    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "get_account_information")
    @mock.patch("snapcraft._store._get_data_from_snap_file")
    def test_sign_build_missing_account_info(
        self, mock_get_snap_data, mock_get_account_info,
    ):
        mock_get_account_info.return_value = {"account_id": "abcd", "snaps": {}}
        mock_get_snap_data.return_value = {"name": "test-snap", "grade": "stable"}

        raised = self.assertRaises(
            storeapi.errors.StoreBuildAssertionPermissionError,
            self.run_command,
            ["sign-build", self.snap_test.snap_path],
        )

        self.assertThat(
            str(raised),
            Equals(
                "Your account lacks permission to assert builds for this "
                "snap. Make sure you are logged in as the publisher of "
                "'test-snap' for series '16'."
            ),
        )

    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "get_account_information")
    @mock.patch("snapcraft._store._get_data_from_snap_file")
    def test_sign_build_no_usable_keys(
        self, mock_get_snap_data, mock_get_account_info,
    ):
        mock_get_account_info.return_value = {
            "account_id": "abcd",
            "snaps": {"16": {"test-snap": {"snap-id": "snap-id"}}},
        }
        mock_get_snap_data.return_value = {"name": "test-snap", "grade": "stable"}

        self.useFixture(
            fixtures.MockPatch("subprocess.check_output", return_value="[]".encode())
        )

        raised = self.assertRaises(
            storeapi.errors.NoKeysError,
            self.run_command,
            ["sign-build", self.snap_test.snap_path],
        )

        self.assertThat(str(raised), Contains("You have no usable keys."))
        self.assertThat(
            str(raised),
            Contains(
                "Please create at least one key with `snapcraft create-key` "
                "for use with snap."
            ),
        )
        snap_build_path = self.snap_test.snap_path + "-build"
        self.assertThat(snap_build_path, Not(FileExists()))

    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "get_account_information")
    @mock.patch("snapcraft._store._get_data_from_snap_file")
    def test_sign_build_no_usable_named_key(
        self, mock_get_snap_data, mock_get_account_info,
    ):
        mock_get_account_info.return_value = {
            "account_id": "abcd",
            "snaps": {"16": {"test-snap": {"snap-id": "snap-id"}}},
        }
        mock_get_snap_data.return_value = {"name": "test-snap", "grade": "stable"}
        self.useFixture(
            fixtures.MockPatch(
                "subprocess.check_output", return_value='[{"name": "default"}]'.encode()
            )
        )

        raised = self.assertRaises(
            storeapi.errors.NoSuchKeyError,
            self.run_command,
            ["sign-build", "--key-name", "zoing", self.snap_test.snap_path],
        )

        self.assertThat(str(raised), Contains("You have no usable key named 'zoing'."))
        self.assertThat(
            str(raised),
            Contains("See the keys available in your system with `snapcraft keys`."),
        )

        snap_build_path = self.snap_test.snap_path + "-build"
        self.assertThat(snap_build_path, Not(FileExists()))

    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "get_account_information")
    @mock.patch("snapcraft._store._get_data_from_snap_file")
    def test_sign_build_unregistered_key(
        self, mock_get_snap_data, mock_get_account_info,
    ):
        mock_get_account_info.return_value = {
            "account_id": "abcd",
            "account_keys": [{"public-key-sha3-384": "another_hash"}],
            "snaps": {"16": {"test-snap": {"snap-id": "snap-id"}}},
        }
        mock_get_snap_data.return_value = {"name": "test-snap", "grade": "stable"}
        self.useFixture(
            fixtures.MockPatch(
                "subprocess.check_output",
                return_value='[{"name": "default", "sha3-384": "a_hash"}]'.encode(),
            )
        )

        raised = self.assertRaises(
            storeapi.errors.KeyNotRegisteredError,
            self.run_command,
            ["sign-build", self.snap_test.snap_path],
        )

        self.assertThat(
            str(raised), Contains("The key 'default' is not registered in the Store.")
        )
        self.assertThat(
            str(raised),
            Contains(
                "Please register it with `snapcraft register-key 'default'` "
                "before signing and uploading signatures to the Store."
            ),
        )
        snap_build_path = self.snap_test.snap_path + "-build"
        self.assertThat(snap_build_path, Not(FileExists()))

    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "get_account_information")
    @mock.patch("snapcraft._store._get_data_from_snap_file")
    def test_sign_build_snapd_failure(
        self, mock_get_snap_data, mock_get_account_info,
    ):
        mock_get_account_info.return_value = {
            "account_id": "abcd",
            "account_keys": [{"public-key-sha3-384": "a_hash"}],
            "snaps": {"16": {"test-snap": {"snap-id": "snap-id"}}},
        }
        mock_get_snap_data.return_value = {"name": "test-snap", "grade": "stable"}
        self.useFixture(
            fixtures.MockPatch(
                "subprocess.check_output",
                side_effect=[
                    '[{"name": "default", "sha3-384": "a_hash"}]'.encode(),
                    subprocess.CalledProcessError(1, ["a", "b"]),
                ],
            )
        )

        raised = self.assertRaises(
            storeapi.errors.SignBuildAssertionError,
            self.run_command,
            ["sign-build", self.snap_test.snap_path],
        )

        self.assertThat(
            str(raised),
            Contains(
                "Failed to sign build assertion for {!r}".format(
                    self.snap_test.snap_path
                )
            ),
        )
        snap_build_path = self.snap_test.snap_path + "-build"
        self.assertThat(snap_build_path, Not(FileExists()))

    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "get_account_information")
    @mock.patch("snapcraft._store._get_data_from_snap_file")
    def test_sign_build_locally_successfully(
        self, mock_get_snap_data, mock_get_account_info,
    ):
        mock_get_account_info.return_value = {
            "account_id": "abcd",
            "snaps": {"16": {"test-snap": {"snap-id": "snap-id"}}},
        }
        mock_get_snap_data.return_value = {"name": "test-snap", "grade": "stable"}
        fake_check_output = fixtures.MockPatch(
            "subprocess.check_output",
            side_effect=['[{"name": "default"}]'.encode(), b"Mocked assertion"],
        )
        self.useFixture(fake_check_output)

        result = self.run_command(["sign-build", self.snap_test.snap_path, "--local"])

        self.assertThat(result.exit_code, Equals(0))
        snap_build_path = self.snap_test.snap_path + "-build"
        self.assertThat(
            result.output,
            Contains("Build assertion {} saved to disk.".format(snap_build_path)),
        )
        self.assertThat(snap_build_path, FileExists())
        fake_check_output.mock.assert_called_with(
            [
                mock.ANY,
                "sign-build",
                "--developer-id=abcd",
                "--snap-id=snap-id",
                "--grade=stable",
                "-k",
                "default",
                self.snap_test.snap_path,
            ]
        )

    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "get_account_information")
    @mock.patch("snapcraft._store._get_data_from_snap_file")
    def test_sign_build_missing_grade(
        self, mock_get_snap_data, mock_get_account_info,
    ):
        mock_get_account_info.return_value = {
            "account_id": "abcd",
            "account_keys": [{"public-key-sha3-384": "a_hash"}],
            "snaps": {"16": {"test-snap": {"snap-id": "snap-id"}}},
        }
        mock_get_snap_data.return_value = {"name": "test-snap"}
        fake_check_output = fixtures.MockPatch(
            "subprocess.check_output",
            side_effect=['[{"name": "default"}]'.encode(), b"Mocked assertion"],
        )
        self.useFixture(fake_check_output)

        result = self.run_command(["sign-build", self.snap_test.snap_path, "--local"])

        self.assertThat(result.exit_code, Equals(0))
        snap_build_path = self.snap_test.snap_path + "-build"
        self.assertThat(
            result.output,
            Contains("Build assertion {} saved to disk.".format(snap_build_path)),
        )
        self.assertThat(snap_build_path, FileExists())
        fake_check_output.mock.assert_called_with(
            [
                mock.ANY,
                "sign-build",
                "--developer-id=abcd",
                "--snap-id=snap-id",
                "--grade=stable",
                "-k",
                "default",
                self.snap_test.snap_path,
            ]
        )

    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "push_snap_build")
    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "get_account_information")
    @mock.patch("snapcraft._store._get_data_from_snap_file")
    def test_sign_build_upload_successfully(
        self, mock_get_snap_data, mock_get_account_info, mock_push_snap_build,
    ):
        mock_get_account_info.return_value = {
            "account_id": "abcd",
            "account_keys": [{"public-key-sha3-384": "a_hash"}],
            "snaps": {"16": {"test-snap": {"snap-id": "snap-id"}}},
        }
        mock_get_snap_data.return_value = {"name": "test-snap", "grade": "stable"}
        fake_check_output = fixtures.MockPatch(
            "subprocess.check_output",
            side_effect=[
                '[{"name": "default", "sha3-384": "a_hash"}]'.encode(),
                b"Mocked assertion",
            ],
        )
        self.useFixture(fake_check_output)

        result = self.run_command(["sign-build", self.snap_test.snap_path])

        self.assertThat(result.exit_code, Equals(0))
        snap_build_path = self.snap_test.snap_path + "-build"
        self.assertThat(
            result.output,
            Contains("Build assertion {} saved to disk.".format(snap_build_path)),
        )
        self.assertThat(
            result.output,
            Contains(
                "Build assertion {} uploaded to the Store.".format(snap_build_path)
            ),
        )
        self.assertThat(snap_build_path, FileExists())
        fake_check_output.mock.assert_called_with(
            [
                mock.ANY,
                "sign-build",
                "--developer-id=abcd",
                "--snap-id=snap-id",
                "--grade=stable",
                "-k",
                "default",
                self.snap_test.snap_path,
            ]
        )
        mock_push_snap_build.assert_called_with("snap-id", "Mocked assertion")

    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "push_snap_build")
    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "get_account_information")
    @mock.patch("snapcraft._store._get_data_from_snap_file")
    def test_sign_build_upload_existing(
        self, mock_get_snap_data, mock_get_account_info, mock_push_snap_build,
    ):
        mock_get_account_info.return_value = {
            "account_id": "abcd",
            "snaps": {"16": {"test-snap": {"snap-id": "snap-id"}}},
        }
        mock_get_snap_data.return_value = {"name": "test-snap", "grade": "stable"}

        snap_build_path = self.snap_test.snap_path + "-build"
        with open(snap_build_path, "wb") as fd:
            fd.write(b"Already signed assertion")

        result = self.run_command(["sign-build", self.snap_test.snap_path])

        self.assertThat(result.exit_code, Equals(0))
        snap_build_path = self.snap_test.snap_path + "-build"
        self.assertThat(
            result.output,
            Contains("A signed build assertion for this snap already exists."),
        )
        self.assertThat(
            result.output,
            Contains(
                "Build assertion {} uploaded to the Store.".format(snap_build_path)
            ),
        )
        mock_push_snap_build.assert_called_with("snap-id", "Already signed assertion")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess
import sys
import urllib.parse
import warnings
from textwrap import dedent
from time import sleep
from typing import Dict, Optional, Sequence

from snapcraft.internal import repo
from snapcraft.internal.errors import SnapcraftEnvironmentError

from .._base_provider import Provider, errors
from ._images import get_image_source

# LXD is only supported on Linux and causes issues when imported on Windows.
# We conditionally import it and rely on ensure_provider() to check OS before
# using pylxd.
if sys.platform == "linux":
    import pylxd

logger = logging.getLogger(__name__)
# Filter out attribute setting warnings for properties that exist in LXD operations
# but are unhandled in pylxd.
warnings.filterwarnings("ignore", module="pylxd.models.operation")


class LXD(Provider):
    """A LXD provider for snapcraft to execute its lifecycle."""

    _PROJECT_DEVICE_NAME = "snapcraft-project"
    _PROJECT_EXPORTED_PRIME_NAME = "snapcraft-project-prime"

    # Given that we are running snapcraft from the snapcraft snap, which has
    # classic confinement and require using the lxd snap, the lxd and lxc
    # binaries should be found in /snap/bin
    _LXD_BIN = os.path.join(os.path.sep, "snap", "bin", "lxd")
    _LXC_BIN = os.path.join(os.path.sep, "snap", "bin", "lxc")

    @classmethod
    def ensure_provider(cls):
        error_message = None  # type: Optional[str]
        prompt_installable = False

        if sys.platform != "linux":
            error_message = "LXD is not supported on this platform"
        else:
            try:
                if not repo.snaps.SnapPackage.is_snap_installed("lxd"):
                    error_message = (
                        "The LXD snap is required to continue: snap install lxd"
                    )
                    prompt_installable = True
            except repo.errors.SnapdConnectionError:
                error_message = (
                    "snap support is required to continue: "
                    "https://docs.snapcraft.io/installing-snapd/6735"
                )

        if error_message is not None:
            raise errors.ProviderNotFound(
                provider=cls._get_provider_name(),
                prompt_installable=prompt_installable,
                error_message=error_message,
            )

        # If we reach this point, it means the lxd snap is properly setup.
        # Now is the time for additional sanity checks to ensure the provider
        # will work.
        try:
            # TODO: add support for more distributions. Maybe refactor a bit so that Repo behaves
            # similar to a build provider.
            if repo.Repo.is_package_installed("lxd") or repo.Repo.is_package_installed(
                "lxd-client"
            ):
                raise SnapcraftEnvironmentError(
                    (
                        "The {!r} provider does not support having the 'lxd' or "
                        "'lxd-client' deb packages installed. To completely migrate "
                        "to the LXD snap run 'lxd.migrate' and try again."
                    ).format(cls._get_provider_name())
                )
        except repo.errors.NoNativeBackendError:
            pass

    @classmethod
    def setup_provider(cls, *, echoer) -> None:
        repo.snaps.install_snaps(["lxd/latest/stable"])

        try:
            subprocess.check_output([cls._LXD_BIN, "waitready", "--timeout=30"])
        except subprocess.CalledProcessError as call_error:
            raise SnapcraftEnvironmentError(
                "Timeout reached waiting for LXD to start."
            ) from call_error

        try:
            subprocess.check_output([cls._LXD_BIN, "init", "--auto"])
        except subprocess.CalledProcessError as call_error:
            raise SnapcraftEnvironmentError(
                "Failed to initialize LXD. "
                "Try manually initializing before trying again: lxd init --auto."
            ) from call_error

    @classmethod
    def _get_provider_name(cls):
        return "LXD"

    @classmethod
    def _get_is_snap_injection_capable(cls) -> bool:
        return True

    def __init__(
        self,
        *,
        project,
        echoer,
        is_ephemeral: bool = False,
        build_provider_flags: Dict[str, str] = None,
    ) -> None:
        super().__init__(
            project=project,
            echoer=echoer,
            is_ephemeral=is_ephemeral,
            build_provider_flags=build_provider_flags,
        )
        # This endpoint is hardcoded everywhere lxc/lxd-pkg-snap#33
        lxd_socket_path = "/var/snap/lxd/common/lxd/unix.socket"
        endpoint = "http+unix://{}".format(urllib.parse.quote(lxd_socket_path, safe=""))
        try:
            self._lxd_client: pylxd.Client = pylxd.Client(endpoint=endpoint)
        except pylxd.client.exceptions.ClientConnectionFailed:
            raise errors.ProviderCommunicationError(
                provider_name=self._get_provider_name(),
                message="cannot connect to the LXD socket ({!r}).".format(
                    lxd_socket_path
                ),
            )

        self._container: Optional[pylxd.models.container.Container] = None

    def _run(
        self, command: Sequence[str], hide_output: bool = False
    ) -> Optional[bytes]:
        self._ensure_container_running()

        env_command = super()._get_env_command()

        # TODO: use pylxd
        cmd = [self._LXC_BIN, "exec", self.instance_name, "--"]
        cmd.extend(env_command)
        cmd.extend(command)
        self._log_run(cmd)

        output = None
        try:
            if hide_output:
                output = subprocess.check_output(cmd)
            else:
                subprocess.check_call(cmd)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderExecError(
                provider_name=self._get_provider_name(),
                command=command,
                exit_code=process_error.returncode,
                output=process_error.output,
            ) from process_error

        return output

    def _launch(self) -> None:
        build_base = self.project._get_build_base()
        try:
            source = get_image_source(base=build_base)
        except KeyError:
            raise errors.ProviderInvalidBaseError(
                provider_name=self._get_provider_name(), build_base=build_base
            )

        config = {"name": self.instance_name, "source": source}

        try:
            container = self._lxd_client.containers.create(config, wait=True)
        except pylxd.exceptions.LXDAPIException as lxd_api_error:
            raise errors.ProviderLaunchError(
                provider_name=self._get_provider_name(), error_message=lxd_api_error
            ) from lxd_api_error
        container.save(wait=True)
        self._container = container

        self._start()

    def _supports_syscall_interception(self) -> bool:
        # syscall interception relies on the seccomp_listener kernel feature
        environment = self._lxd_client.host_info.get("environment", {})
        kernel_features = environment.get("kernel_features", {})
        return kernel_features.get("seccomp_listener", "false") == "true"

    def _start(self):
        if not self._lxd_client.containers.exists(self.instance_name):
            raise errors.ProviderInstanceNotFoundError(instance_name=self.instance_name)

        if self._container is None:
            self._container = self._lxd_client.containers.get(self.instance_name)

        self._container.sync()

        # map to the owner of the directory we are eventually going to write the
        # snap to.
        self._container.config["raw.idmap"] = "both {!s} 0".format(
            os.stat(self.project._project_dir).st_uid
        )
        # If possible, allow container to make safe mknod calls. Documented at
        # https://linuxcontainers.org/lxd/docs/master/syscall-interception
        if self._supports_syscall_interception():
            self._container.config["security.syscalls.intercept.mknod"] = "true"
        self._container.save(wait=True)

        if self._container.status.lower() != "running":
            try:
                self._container.start(wait=True)
            except pylxd.exceptions.LXDAPIException as lxd_api_error:
                print(self._container.status)
                raise errors.ProviderStartError(
                    provider_name=self._get_provider_name(), error_message=lxd_api_error
                ) from lxd_api_error

        self.echoer.wrapped("Waiting for container to be ready")

    def _stop(self):
        # If _container is still None here it means creation/starting was not
        # successful.
        if self._container is None:
            return

        self._container.sync()

        if self._container.status.lower() != "stopped":
            try:
                self._container.stop(wait=True)
            except pylxd.exceptions.LXDAPIException as lxd_api_error:
                raise errors.ProviderStopError(
                    provider_name=self._get_provider_name(), error_message=lxd_api_error
                ) from lxd_api_error

    def _push_file(self, *, source: str, destination: str) -> None:
        # Sanity check - developer error if container not initialized.
        if self._container is None:
            raise RuntimeError("Attempted to use container before starting.")

        self._ensure_container_running()

        # TODO: better handling of larger files.
        with open(source, "rb") as source_data:
            source_contents = source_data.read()

        try:
            self._container.files.put(destination, source_contents)
        except pylxd.exceptions.LXDAPIException as lxd_api_error:
            raise errors.ProviderFileCopyError(
                provider_name=self._get_provider_name(), error_message=lxd_api_error
            )

    def create(self) -> None:
        """Create the LXD instance and setup the build environment."""
        self.echoer.info("Launching a container.")
        self.launch_instance()
        self._mount_project()

    def destroy(self) -> None:
        """Destroy the instance, trying to stop it first."""
        self._stop()

    def _get_mount_name(self, target: str) -> str:
        """Provide a formatted name for target mount point."""
        home_dir = self._get_home_directory().as_posix()

        # Special cases for compatibility.
        if target == os.path.join(home_dir, "project"):
            return self._PROJECT_DEVICE_NAME
        elif target == os.path.join(home_dir, "prime"):
            return self._PROJECT_EXPORTED_PRIME_NAME

        # Replace home directory with "snapcraft".
        name = target.replace(home_dir, "snapcraft", 1)

        # Replace path separators with dashes.
        name = name.replace("/", "-")
        return name

    def _is_mounted(self, target: str) -> bool:
        """Query if there is a mount at target mount point."""
        # Sanity check - developer error if container not initialized.
        if self._container is None:
            raise RuntimeError("Attempted to use container before starting.")

        name = self._get_mount_name(target)
        return name in self._container.devices

    def _mount(self, host_source: str, target: str) -> None:
        """Mount host source directory to target mount point."""
        # Sanity check - developer error if container not initialized.
        if self._container is None:
            raise RuntimeError("Attempted to use container before starting.")

        if self._is_mounted(target):
            # Nothing to do if already mounted.
            return

        name = self._get_mount_name(target)
        self._container.sync()
        self._container.devices[name] = {
            "type": "disk",
            "source": host_source,
            "path": target,
        }

        try:
            self._container.save(wait=True)
        except pylxd.exceptions.LXDAPIException as lxd_api_error:
            raise errors.ProviderMountError(
                provider_name=self._get_provider_name(), error_message=lxd_api_error
            ) from lxd_api_error

    def clean_project(self) -> bool:
        was_cleaned = super().clean_project()
        if was_cleaned:
            if self._container is None:
                try:
                    self._container = self._lxd_client.containers.get(
                        self.instance_name
                    )
                except pylxd.exceptions.NotFound:
                    # If no container found, nothing to delete.
                    return was_cleaned
            self._stop()
            self._container.delete(wait=True)
        return was_cleaned

    def pull_file(self, name: str, destination: str, delete: bool = False) -> None:
        # Sanity check - developer error if container not initialized.
        if self._container is None:
            raise RuntimeError("Attempted to use container before starting.")

        self._ensure_container_running()

        # TODO: better handling of larger files.
        try:
            source_data = self._container.files.get(name)
        except pylxd.exceptions.LXDAPIException as lxd_api_error:
            raise errors.ProviderFileCopyError(
                provider_name=self._get_provider_name(), error_message=lxd_api_error
            )
        else:
            with open(destination, "wb") as destination_data:
                destination_data.write(source_data)
        if delete and self._container.files.delete_available:
            self._container.files.delete(name)
        if delete and not self._container.files.delete_available:
            logger.warning("File deletion not supported by this LXD version.")

    def shell(self) -> None:
        self._run(command=["/bin/bash"])

    def _wait_for_systemd(self) -> None:
        # systemctl states we care about here are:
        # - running: The system is fully operational. Process returncode: 0
        # - degraded: The system is operational but one or more units failed.
        #             Process returncode: 1
        for i in range(40):
            try:
                self._run(["systemctl", "is-system-running"], hide_output=True)
                break
            except errors.ProviderExecError as exec_error:
                if exec_error.output is not None:
                    running_state = exec_error.output.decode().strip()
                    if running_state == "degraded":
                        break
                    logger.debug(f"systemctl is-system-running: {running_state!r}")
                sleep(0.5)
        else:
            self.echoer.warning("Timed out waiting for systemd to be ready...")

    def _wait_for_network(self) -> None:
        self.echoer.wrapped("Waiting for network to be ready...")
        for i in range(40):
            try:
                self._run(["getent", "hosts", "snapcraft.io"], hide_output=True)
                break
            except errors.ProviderExecError:
                sleep(0.5)
        else:
            self.echoer.warning("Failed to setup networking.")

    def _setup_environment(self) -> None:
        if self._container is None:
            raise RuntimeError("Attempted to use container before starting.")

        super()._setup_environment()

        self._install_file(
            path="/etc/systemd/network/10-eth0.network",
            content=dedent(
                """
                [Match]
                Name=eth0

                [Network]
                DHCP=ipv4
                LinkLocalAddressing=ipv6

                [DHCP]
                RouteMetric=100
                UseMTU=true
                """
            ),
            permissions="0644",
        )

        self._install_file(
            path="/etc/hostname", content=self.instance_name, permissions="0644"
        )
        self._run(
            ["hostname", "-F", "/etc/hostname"], hide_output=True,
        )

        self._wait_for_systemd()

        # Use resolv.conf managed by systemd-resolved.
        self._run(
            ["ln", "-sf", "/run/systemd/resolve/resolv.conf", "/etc/resolv.conf"],
            hide_output=True,
        )

        self._run(["systemctl", "enable", "systemd-resolved"], hide_output=True)
        self._run(["systemctl", "enable", "systemd-networkd"], hide_output=True)

        self._run(["systemctl", "restart", "systemd-resolved"], hide_output=True)
        self._run(["systemctl", "restart", "systemd-networkd"], hide_output=True)

        self._wait_for_network()

        # Setup snapd to bootstrap.
        self._run(["apt-get", "update"])

        # First install fuse and udev, snapd requires them.
        # Snapcraft requires dirmngr
        self._run(["apt-get", "install", "dirmngr", "udev", "fuse", "--yes"])

        # the system needs networking
        self._run(["systemctl", "enable", "systemd-udevd"], hide_output=True)
        self._run(["systemctl", "start", "systemd-udevd"], hide_output=True)

        # And only then install snapd.
        self._run(["apt-get", "install", "snapd", "sudo", "--yes"])
        self._run(["systemctl", "start", "snapd"], hide_output=True)

    def _setup_snapcraft(self):
        self._wait_for_network()
        super()._setup_snapcraft()

    def _ensure_container_running(self) -> None:
        # Sanity check - developer error if container not initialized.
        if self._container is None:
            raise RuntimeError("Attempted to use container before starting.")

        self._container.sync()
        if self._container.status.lower() != "running":
            raise errors.ProviderFileCopyError(
                provider_name=self._get_provider_name(),
                error_message=(
                    "Container is not running, the current state is: {!r}. "
                    "Ensure it has not been modified by external factors and try again"
                ).format(self._container.status),
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

import attr


@attr.s(auto_attribs=True)
class DebPackage(object):
    """Debian package representation."""

    name: str
    arch: Optional[str] = None
    version: Optional[str] = None

    @classmethod
    def from_unparsed(cls, package: str) -> "DebPackage":
        """Parse package supported in yaml.

        Package Format:
        <package-name>[:<arch>][=<version>]

        Examples:
        "foo"
        "foo:i386"
        "foo=1.5"
        "foo:i386=1.5"

        :param package: Package to parse.

        :returns: DebPackage with populated arch & version, if any.
        """
        parsed_name: str = package
        parsed_arch: Optional[str] = None
        parsed_version: Optional[str] = None

        if "=" in parsed_name:
            parsed_name, parsed_version = parsed_name.split("=")

        if ":" in parsed_name:
            parsed_name, parsed_arch = parsed_name.split(":")

        return DebPackage(name=parsed_name, arch=parsed_arch, version=parsed_version)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.states._state import PartState


class _TestState(PartState):
    def properties_of_interest(self, part_properties):
        return {"foo": part_properties.get("foo"), "baz": part_properties.get("baz")}

    def project_options_of_interest(self, project):
        return {
            "foo": getattr(project, "foo", None),
            "baz": getattr(project, "baz", None),
        }


class _TestProject:
    def __init__(self, dict):
        for key, value in dict.items():
            setattr(self, key, value)


class TestState:
    scenarios = [
        ("change existing", dict(old={"foo": "bar"}, new={"foo": "baz"})),
        ("add new", dict(old={"baz": "qux"}, new={"foo": "bar", "baz": "qux"})),
        ("remove old", dict(old={"foo": "bar", "baz": "qux"}, new={"baz": "qux"})),
    ]

    def get_state(self, old):
        project = _TestProject(old)
        part_properties = old

        state = _TestState(part_properties, project)
        state.properties = old
        state.project_options = old

        return state

    def test_diff_properties_of_interest(self, old, new):
        state = self.get_state(old)
        differing_properties = state.diff_properties_of_interest(new)

        assert differing_properties == {"foo"}

    def test_diff_project_options_of_interest(self, old, new):
        state = self.get_state(old)
        differing_properties = state.diff_project_options_of_interest(_TestProject(new))

        assert differing_properties == {"foo"}



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re
import shutil
from contextlib import ContextDecorator
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

import apt

from snapcraft.internal import common
from snapcraft.internal.indicators import is_dumb_terminal
from snapcraft.internal.repo import errors
from snapcraft.internal.repo._base import get_pkg_name_parts

logger = logging.getLogger(__name__)


_HASHSUM_MISMATCH_PATTERN = re.compile(r"(E:Failed to fetch.+Hash Sum mismatch)+")


class AptCache(ContextDecorator):
    """Transient cache for use with stage-packages, or read-only host-mode for build-packages."""

    def __init__(
        self,
        *,
        stage_cache: Optional[Path] = None,
        stage_cache_arch: Optional[str] = None,
    ) -> None:
        self.stage_cache = stage_cache
        self.stage_cache_arch = stage_cache_arch

    def __enter__(self) -> "AptCache":
        if self.stage_cache is not None:
            self._configure_apt()
            self._populate_stage_cache_dir()
            self.cache = apt.Cache(rootdir=str(self.stage_cache), memonly=True)
        else:
            # There appears to be a slowdown when using `rootdir` = '/' with
            # apt.Cache().  Do not set it for the host cache.
            self.cache = apt.Cache()
        return self

    def __exit__(self, *exc) -> None:
        self.cache.close()
        self.cache = None

    def _configure_apt(self):
        # Do not install recommends.
        apt.apt_pkg.config.set("Apt::Install-Recommends", "False")

        # Ensure repos are provided by trusted third-parties.
        apt.apt_pkg.config.set("Acquire::AllowInsecureRepositories", "False")

        # Methods and solvers dir for when in the SNAP.
        snap_dir = os.getenv("SNAP")
        if common.is_snap() and snap_dir and os.path.exists(snap_dir):
            apt_dir = os.path.join(snap_dir, "usr", "lib", "apt")
            apt.apt_pkg.config.set("Dir", apt_dir)
            # yes apt is broken like that we need to append os.path.sep
            methods_dir = os.path.join(apt_dir, "methods")
            apt.apt_pkg.config.set("Dir::Bin::methods", methods_dir + os.path.sep)
            solvers_dir = os.path.join(apt_dir, "solvers")
            apt.apt_pkg.config.set("Dir::Bin::solvers::", solvers_dir + os.path.sep)
            apt_key_path = os.path.join(snap_dir, "usr", "bin", "apt-key")
            apt.apt_pkg.config.set("Dir::Bin::apt-key", apt_key_path)
            gpgv_path = os.path.join(snap_dir, "usr", "bin", "gpgv")
            apt.apt_pkg.config.set("Apt::Key::gpgvcommand", gpgv_path)

        apt.apt_pkg.config.set("Dir::Etc::Trusted", "/etc/apt/trusted.gpg")
        apt.apt_pkg.config.set("Dir::Etc::TrustedParts", "/etc/apt/trusted.gpg.d/")

        # Clear up apt's Post-Invoke-Success as we are not running
        # on the system.
        apt.apt_pkg.config.clear("APT::Update::Post-Invoke-Success")

        self.progress = apt.progress.text.AcquireProgress()
        if is_dumb_terminal():
            # Make output more suitable for logging.
            self.progress.pulse = lambda owner: True
            self.progress._width = 0

    def _populate_stage_cache_dir(self) -> None:
        """Create/refresh cache configuration.

        (1) Delete old-style symlink cache, if symlink.
        (2) Delete current-style (copied) tree.
        (3) Copy current host apt configuration.
        (4) Configure primary arch to target arch.
        (5) Install dpkg into cache directory to support multi-arch.
        """
        if self.stage_cache is None:
            return

        # Copy apt configuration from host.
        cache_etc_apt_path = Path(self.stage_cache, "etc", "apt")

        # Delete potentially outdated cache configuration.
        if cache_etc_apt_path.is_symlink():
            cache_etc_apt_path.unlink()
        elif cache_etc_apt_path.exists():
            shutil.rmtree(cache_etc_apt_path)

        # Copy current cache configuration.
        cache_etc_apt_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copytree("/etc/apt", cache_etc_apt_path)

        # Specify default arch (if specified).
        if self.stage_cache_arch is not None:
            arch_conf_path = cache_etc_apt_path / "apt.conf.d" / "00default-arch"
            arch_conf_path.write_text(f'APT::Architecture "{self.stage_cache_arch}";\n')

        # dpkg also needs to be in the rootdir in order to support multiarch
        # (apt calls dpkg --print-foreign-architectures).
        dpkg_path = shutil.which("dpkg")
        if dpkg_path:
            # Symlink it into place
            destination = Path(self.stage_cache, dpkg_path[1:])
            if not destination.exists():
                destination.parent.mkdir(parents=True, exist_ok=True)
                os.symlink(dpkg_path, destination)
        else:
            logger.warning("Cannot find 'dpkg' command needed to support multiarch")

    def _autokeep_packages(self) -> None:
        # If the package has been installed automatically as a dependency
        # of another package, and if no packages depend on it anymore,
        # the package is no longer required.
        for package in self.cache.get_changes():
            if package.is_auto_removable:
                package.mark_keep()

    def _set_pkg_version(self, package: apt.Package, version: str) -> None:
        # Set candidate version to a specific version if available
        if version in package.versions:
            version = package.versions.get(version)
            package.candidate = version
        else:
            raise errors.PackageNotFoundError("{}={}".format(package.name, version))

    def _verify_marked_install(self, package: apt.Package):
        if not package.installed and not package.marked_install:
            broken_deps: List[str] = list()

            if package.candidate is None:
                raise errors.PackageNotFoundError(package.name)

            for package_dependencies in package.candidate.dependencies:
                for dep in package_dependencies:
                    if not dep.target_versions:
                        broken_deps.append(dep.name)
            raise errors.PackageBrokenError(package.name, broken_deps)

    def is_package_valid(self, package_name: str) -> bool:
        return package_name in self.cache or self.cache.is_virtual_package(package_name)

    def get_installed_version(
        self, package_name: str, *, resolve_virtual_packages: bool = False
    ) -> Optional[str]:
        if resolve_virtual_packages and self.cache.is_virtual_package(package_name):
            logger.warning(
                f"{package_name!r} is a virtual package, use non virtual packages for deterministic results."
            )
            # Recusrse until a "real" package is found.
            return self.get_installed_version(
                self.cache.get_providing_packages(package_name)[0].name,
                resolve_virtual_packages=resolve_virtual_packages,
            )

        if package_name in self.cache:
            if self.cache[package_name].installed is not None:
                return self.cache[package_name].installed.version
        return None

    def fetch_archives(self, download_path: Path) -> List[Tuple[str, str, Path]]:
        """Fetches archives, list of (<package-name>, <package-version>, <dl-path>)."""
        downloaded = list()
        for package in self.cache.get_changes():
            try:
                dl_path = package.candidate.fetch_binary(str(download_path))
            except apt.package.FetchError as e:
                raise errors.PackageFetchError(str(e))

            if package.candidate is None:
                raise errors.PackageNotFoundError(package.name)

            downloaded.append((package.name, package.candidate.version, Path(dl_path)))
        return downloaded

    def get_installed_packages(self) -> Dict[str, str]:
        installed: Dict[str, str] = dict()
        for package in self.cache:
            if package.installed is not None:
                installed[package.name] = str(package.installed.version)
        return installed

    def get_packages_marked_for_installation(self) -> List[Tuple[str, str]]:
        changed_packages = self.cache.get_changes()
        marked_install_packages = [p for p in changed_packages if p.marked_install]
        package_names_missing_installation_candidate = [
            p.name for p in marked_install_packages if p.candidate is None
        ]

        if package_names_missing_installation_candidate:
            raise errors.PackagesNotFoundError(
                package_names_missing_installation_candidate
            )

        return [(p.name, p.candidate.version) for p in marked_install_packages]

    def mark_packages(self, package_names: Set[str]) -> None:
        for name in package_names:
            if name.endswith(":any"):
                name = name[:-4]

            if self.cache.is_virtual_package(name):
                name = self.cache.get_providing_packages(name)[0].name

            logger.debug(
                "Marking {!r} (and its dependencies) to be fetched".format(name)
            )

            name_arch, version = get_pkg_name_parts(name)
            if name_arch not in self.cache:
                raise errors.PackageNotFoundError(name_arch)

            package = self.cache[name_arch]
            if version is not None:
                self._set_pkg_version(package, version)

            logger.debug(f"package: {package!r}")

            # Disable automatic resolving of broken packages here
            # because if that fails it raises a SystemError and the
            # API doesn't expose enough information about the problem.
            # Instead we let apt-get show a verbose error message later.
            # Also, make sure this package is marked as auto-installed,
            # which will propagate to its dependencies.
            package.mark_install(auto_fix=False, from_user=False)

            # Now mark this package as NOT automatically installed, which
            # will leave its dependencies marked as auto-installed, which
            # allows us to clean them up if necessary.
            package.mark_auto(False)

            self._verify_marked_install(package)

    def unmark_packages(self, unmark_names: Set[str]) -> None:
        skipped_essential = set()
        skipped_filtered = set()

        for package in self.cache.get_changes():
            if package.candidate is None:
                raise errors.PackageNotFoundError(package.name)

            if package.candidate.priority == "essential":
                # Filter 'essential' packages.
                skipped_essential.add(package.name)
                package.mark_keep()
                continue

            if package.name in unmark_names:
                # Filter packages from given list.
                skipped_filtered.add(package.name)
                package.mark_keep()
                continue

        if skipped_essential:
            logger.debug(
                f"Skipping priority essential packages: {sorted(skipped_essential)}"
            )

        if skipped_filtered:
            logger.debug(
                f"Skipping filtered manifest packages: {sorted(skipped_filtered)}"
            )

        # Unmark dependencies that are no longer required.
        self._autokeep_packages()

    def update(self) -> None:
        try:
            self.cache.update(fetch_progress=self.progress, sources_list=None)
            self.cache.close()
            self.cache = apt.Cache(rootdir=str(self.stage_cache), memonly=True)
        except apt.cache.FetchFailedException as e:
            raise errors.CacheUpdateFailedError(str(e))



#!/usr/bin/env python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os
import subprocess
import sys

DEFAULT_REFERENCE_FILE = "docs/reference.md"


def _get_help_output(topic, lines=False):
    TRANSLATE = {"tar-content": "tar_content"}
    if topic in TRANSLATE:
        topic = TRANSLATE[topic]
    return _run_snapcraft(["help", topic], lines)


def _run_snapcraft(cmd, lines=False):
    cmd = ["./snapcraft"] + cmd
    output = subprocess.check_output(
        cmd, cwd=os.path.join(os.path.dirname(__file__), "../../bin")
    )
    output = output.decode("utf-8")
    if lines:
        return output.split("\n")[:-1]  # drop last empty line
    return output


class Reference:
    def __init__(self, output_file):
        self.docs = []
        self.topics = []
        self.plugins = []
        self.output_file = output_file
        assert os.path.isdir(
            os.path.dirname(self.output_file)
        ), "Directory {} does not exist.".format(os.path.dirname(self.output_file))
        self._read()
        self._write()

    def _read(self):
        self.topics = _get_help_output("topics", lines=True)
        self.plugins = _run_snapcraft(["list-plugins"], lines=True)
        self.topics.extend(self.plugins)
        for topic in self.topics:
            text = _get_help_output(topic)
            if not text.startswith("The plugin has no documentation"):
                self.docs += [(topic, text)]

    def _write(self):
        reference_text = """
# Snapcraft reference

`snapcraft` comes with a handy help system, just run:
```
$ snapcraft help
```
to get you started.

Here we want to provide you with the current documentation from within
snappy. It's a reference of `snapcraft`'s plugins and internals.

## Internal documentation
"""
        for entry in [e for e in self.docs if e[0] not in self.plugins]:
            reference_text += """
### Topic: {}

{}""".format(
                entry[0].title(), entry[1].replace("# ", "#### ")
            )
        reference_text += """

## snapcraft's Plugins

In this section we are going to discuss `snapcraft`'s plugins, their workflow
and their options.
"""
        for entry in [e for e in self.docs if e[0] in self.plugins]:
            reference_text += """
### The {} plugin

{}""".format(
                entry[0], entry[1]
            )
        with open(os.path.join(self.output_file), "w") as f:
            f.write(str(reference_text))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-O",
        "--output-file",
        help="specify output file, default: {}".format(DEFAULT_REFERENCE_FILE),
        default=DEFAULT_REFERENCE_FILE,
    )
    args = parser.parse_args()
    Reference(args.output_file)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("Aborted.")
        sys.exit(1)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from textwrap import dedent
from unittest import mock
from unittest.mock import call

import fixtures
import pytest
from testtools.matchers import Equals

from snapcraft.internal import steps
from snapcraft.internal.build_providers import _base_provider, errors
from snapcraft.internal.build_providers._multipass import Multipass, MultipassCommand
from snapcraft.internal.errors import SnapcraftEnvironmentError
from tests.unit.build_providers import BaseProviderBaseTest, get_project

_DEFAULT_INSTANCE_INFO = dedent(
    """\
    {
        "errors": [
        ],
        "info": {
            "snapcraft-project-name": {
                "disks": {
                    "sda1": {
                        "total": "5136297984",
                        "used": "1076072448"
                    }
                },
                "image_hash": "bc990872f070249450fc187d668bf65e2f87ffa3d7cf6e2934cb84b62af368e1",
                "image_release": "16.04 LTS",
                "ipv4": [
                    "10.74.70.141"
                ],
                "load": [
                    0,
                    0,
                    0
                ],
                "memory": {
                    "total": 1040318464,
                    "used": 42754048
                },
                "mounts": {
                },
                "release": "Ubuntu 16.04.4 LTS",
                "state": "RUNNING"
            }
        }
    }
    """
)  # noqa: E501


@pytest.fixture()
def multipass_cmd():
    """Fake MultipassCommand implementation."""

    def execute_effect(*, command, instance_name, hide_output):
        if hide_output:
            return None
        return b""

    patcher = mock.patch(
        "snapcraft.internal.build_providers._multipass." "_multipass.MultipassCommand",
        spec=MultipassCommand,
    )
    multipass_cmd_mock = patcher.start()

    multipass_cmd_mock().execute.side_effect = execute_effect

    # default data returned for info so launch is triggered
    multipass_cmd_mock().info.side_effect = [
        errors.ProviderInfoError(
            provider_name="multipass", exit_code=1, stderr=b"error"
        ),
        _DEFAULT_INSTANCE_INFO.encode(),
        _DEFAULT_INSTANCE_INFO.encode(),
    ]
    multipass_cmd_mock().exists.return_value = False

    yield multipass_cmd_mock

    patcher.stop()


class GetEnv(_base_provider.Provider):
    def _get_env_command(self):
        return ["env", "SNAPCRAFT_HAS_TTY=False"]


class MultipassTestImpl(Multipass, GetEnv):
    pass


class MultipassTest(BaseProviderBaseTest):
    def setUp(self):
        super().setUp()

        patcher = mock.patch(
            "snapcraft.internal.build_providers._multipass."
            "_multipass.MultipassCommand",
            spec=MultipassCommand,
        )
        self.multipass_cmd_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch(
            "snapcraft.internal.build_providers._base_provider.Provider.clean_project",
            return_value=True,
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("builtins.open", mock.mock_open())
        self.open_mock = patcher.start()
        self.addCleanup(patcher.stop)

        # default data returned for info so launch is triggered
        self.multipass_cmd_mock().info.side_effect = [
            errors.ProviderInfoError(
                provider_name="multipass", exit_code=1, stderr=b"error"
            ),
            _DEFAULT_INSTANCE_INFO.encode(),
            _DEFAULT_INSTANCE_INFO.encode(),
        ]

    def test_ephemeral_instance_with_contextmanager(self):
        with MultipassTestImpl(
            project=self.project, echoer=self.echoer_mock, is_ephemeral=True
        ) as instance:
            instance.execute_step(steps.PULL)
            instance.execute_step(steps.BUILD)

        self.multipass_cmd_mock().launch.assert_called_once_with(
            instance_name=self.instance_name,
            cpus="2",
            mem="2G",
            disk="256G",
            image="snapcraft:core16",
        )
        # Given SnapInjector is mocked, we only need to verify the commands
        # called from the Multipass class.
        self.multipass_cmd_mock().execute.assert_has_calls(
            [
                mock.call(
                    instance_name=self.instance_name,
                    hide_output=False,
                    command=[
                        "sudo",
                        "-H",
                        "-i",
                        "env",
                        "SNAPCRAFT_HAS_TTY=False",
                        "snapcraft",
                        "pull",
                    ],
                ),
                mock.call(
                    instance_name=self.instance_name,
                    hide_output=False,
                    command=[
                        "sudo",
                        "-H",
                        "-i",
                        "env",
                        "SNAPCRAFT_HAS_TTY=False",
                        "snapcraft",
                        "build",
                    ],
                ),
            ]
        )
        self.assertThat(self.multipass_cmd_mock().info.call_count, Equals(3))
        self.multipass_cmd_mock().info.assert_has_calls(
            [
                mock.call(instance_name=self.instance_name, output_format="json"),
                mock.call(instance_name=self.instance_name, output_format="json"),
            ]
        )
        self.multipass_cmd_mock().stop.assert_called_once_with(
            instance_name=self.instance_name, time=10
        )
        self.multipass_cmd_mock().delete.assert_called_once_with(
            instance_name=self.instance_name, purge=True
        )

    def test_launch_with_cpus_from_environment(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT_CPU", "64")
        )

        instance = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)
        instance.create()

        self.multipass_cmd_mock().launch.assert_called_once_with(
            instance_name=self.instance_name,
            cpus="64",
            mem="2G",
            disk="256G",
            image="snapcraft:core16",
        )

    def test_launch_for_type_base(self):
        self.project._snap_meta.name = "core18"
        self.project._snap_meta.type = "base"
        self.project._snap_meta.base = None

        instance = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)
        self.useFixture(
            fixtures.MockPatchObject(
                instance,
                "_get_instance_info",
                side_effect=[
                    errors.ProviderInfoError(
                        provider_name="multipass", exit_code=1, stderr=b"error"
                    ),
                    {},
                ],
            )
        )
        self.useFixture(fixtures.MockPatchObject(instance, "_mount_project",))

        instance.create()

        self.multipass_cmd_mock().launch.assert_called_once_with(
            instance_name="snapcraft-core18",
            cpus="2",
            mem="2G",
            disk="256G",
            image="snapcraft:core18",
        )

    def test_launch_with_ram_from_environment(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT_MEMORY", "4G")
        )

        instance = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)
        instance.create()

        self.multipass_cmd_mock().launch.assert_called_once_with(
            instance_name=self.instance_name,
            cpus="2",
            mem="4G",
            disk="256G",
            image="snapcraft:core16",
        )

    def test_launch_with_disk_from_environment(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT_DISK", "400G")
        )

        instance = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)
        instance.create()

        self.multipass_cmd_mock().launch.assert_called_once_with(
            instance_name=self.instance_name,
            cpus="2",
            mem="2G",
            disk="400G",
            image="snapcraft:core16",
        )

    def test_push_file(self):
        multipass = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)

        multipass._push_file(source="src.txt", destination="dest.txt")

        self.multipass_cmd_mock().push_file.assert_called_once_with(
            destination="{}:dest.txt".format(self.instance_name),
            source=self.open_mock.return_value.__enter__(),
        )

        self.open_mock.assert_called_once_with("src.txt", "rb")
        self.open_mock.return_value.__exit__.assert_called()
        self.multipass_cmd_mock().info.assert_not_called()
        self.multipass_cmd_mock().stop.assert_not_called()
        self.multipass_cmd_mock().delete.assert_not_called()

    def test_pull_file(self):
        multipass = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)

        multipass.pull_file("src.txt", "dest.txt")

        self.multipass_cmd_mock().execute.assert_called_once_with(
            command=[
                "sudo",
                "-H",
                "-i",
                "env",
                "SNAPCRAFT_HAS_TTY=False",
                "test",
                "-f",
                "src.txt",
            ],
            hide_output=False,
            instance_name="snapcraft-project-name",
        )

        self.multipass_cmd_mock().pull_file.assert_called_once_with(
            destination=self.open_mock.return_value.__enter__(),
            source="{}:src.txt".format(self.instance_name),
        )

        self.open_mock.assert_called_once_with("dest.txt", "wb")
        self.open_mock.return_value.__exit__.assert_called()
        self.multipass_cmd_mock().info.assert_not_called()
        self.multipass_cmd_mock().stop.assert_not_called()
        self.multipass_cmd_mock().delete.assert_not_called()

    def test_pull_and_delete_file(self):
        multipass = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)

        multipass.pull_file("src.txt", "dest.txt", delete=True)

        self.multipass_cmd_mock().pull_file.assert_called_once_with(
            destination=self.open_mock.return_value.__enter__(),
            source="{}:src.txt".format(self.instance_name),
        )

        self.multipass_cmd_mock().execute.assert_has_calls(
            [
                mock.call(
                    command=[
                        "sudo",
                        "-H",
                        "-i",
                        "env",
                        "SNAPCRAFT_HAS_TTY=False",
                        "test",
                        "-f",
                        "src.txt",
                    ],
                    hide_output=False,
                    instance_name="snapcraft-project-name",
                ),
                mock.call(
                    command=[
                        "sudo",
                        "-H",
                        "-i",
                        "env",
                        "SNAPCRAFT_HAS_TTY=False",
                        "rm",
                        "src.txt",
                    ],
                    hide_output=False,
                    instance_name="snapcraft-project-name",
                ),
            ]
        )
        self.open_mock.assert_called_once_with("dest.txt", "wb")
        self.open_mock.return_value.__exit__.assert_called()
        self.multipass_cmd_mock().info.assert_not_called()
        self.multipass_cmd_mock().stop.assert_not_called()
        self.multipass_cmd_mock().delete.assert_not_called()

    def test_instance_does_not_exist_on_destroy(self):
        # An error is raised if the queried image does not exist
        self.multipass_cmd_mock().info.side_effect = errors.ProviderInfoError(
            provider_name=self.instance_name, exit_code=2, stderr=b"error"
        )

        multipass = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)

        multipass.destroy()

        self.multipass_cmd_mock().stop.assert_not_called()
        self.multipass_cmd_mock().delete.assert_not_called()

    def test_destroy_instance_with_stop_delay(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT_STOP_TIME", "60")
        )

        multipass = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)

        multipass.create()
        multipass.destroy()

        self.multipass_cmd_mock().stop.assert_called_once_with(
            instance_name=self.instance_name, time=60
        )
        self.multipass_cmd_mock().delete.assert_not_called()

    def test_destroy_instance_with_stop_delay_0(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT_STOP_TIME", "0")
        )

        multipass = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)

        multipass.create()
        multipass.destroy()

        self.multipass_cmd_mock().stop.assert_called_once_with(
            instance_name=self.instance_name
        )
        self.multipass_cmd_mock().delete.assert_not_called()

    def test_destroy_instance_with_stop_delay_invalid(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT_STOP_TIME", "A")
        )

        multipass = MultipassTestImpl(project=self.project, echoer=self.echoer_mock)

        multipass.create()

        self.assertRaises(SnapcraftEnvironmentError, multipass.destroy)

        self.multipass_cmd_mock().stop.assert_not_called()
        self.multipass_cmd_mock().delete.assert_not_called()


class TestMultipassWithBases:

    scenarios = (
        ("linux", dict(base="core16", expected_image="snapcraft:core16")),
        ("linux", dict(base="core18", expected_image="snapcraft:core18")),
    )

    def test_lifecycle(
        self, xdg_dirs, in_snap, snap_injector, multipass_cmd, base, expected_image
    ):
        with MultipassTestImpl(
            project=get_project(base=base), echoer=mock.Mock(), is_ephemeral=False
        ) as instance:
            instance.execute_step(steps.PULL)

        multipass_cmd().launch.assert_called_once_with(
            instance_name="snapcraft-project-name",
            cpus="2",
            mem="2G",
            disk="256G",
            image=expected_image,
        )

        multipass_cmd().execute.assert_called()

        for args, kwargs in multipass_cmd().execute.call_args_list:
            assert kwargs["command"][0:3] == [
                "sudo",
                "-H",
                "-i",
            ]
            assert kwargs["hide_output"] is False
            assert kwargs["instance_name"] == "snapcraft-project-name"

        assert multipass_cmd().mount.mock_calls == [
            mock.call(
                source=mock.ANY,
                target="snapcraft-project-name:/root/project",
                uid_map={str(os.getuid()): "0"},
                gid_map={str(os.getgid()): "0"},
            )
        ]
        multipass_cmd().umount.assert_not_called()

        assert multipass_cmd().info.call_count == 3
        multipass_cmd().info.assert_has_calls(
            [
                mock.call(instance_name="snapcraft-project-name", output_format="json"),
                mock.call(instance_name="snapcraft-project-name", output_format="json"),
            ]
        )
        assert multipass_cmd().push_file.mock_calls == [
            call(
                destination="snapcraft-project-name:/var/tmp/L3Jvb3QvLmJhc2hyYw==",
                source=mock.ANY,
            ),
            call(
                destination="snapcraft-project-name:/var/tmp/L2Jpbi9fc25hcGNyYWZ0X3Byb21wdA==",
                source=mock.ANY,
            ),
            call(
                destination="snapcraft-project-name:/var/tmp/L2V0Yy9hcHQvc291cmNlcy5saXN0",
                source=mock.ANY,
            ),
            call(
                destination="snapcraft-project-name:/var/tmp/L2V0Yy9hcHQvc291cmNlcy5saXN0LmQvZGVmYXVsdC5zb3VyY2Vz",
                source=mock.ANY,
            ),
            call(
                destination="snapcraft-project-name:/var/tmp/L2V0Yy9hcHQvc291cmNlcy5saXN0LmQvZGVmYXVsdC1zZWN1cml0eS5zb3VyY2Vz",
                source=mock.ANY,
            ),
            call(
                destination="snapcraft-project-name:/var/tmp/L2V0Yy9hcHQvYXB0LmNvbmYuZC8wMC1zbmFwY3JhZnQ=",
                source=mock.ANY,
            ),
        ]

        multipass_cmd().stop.assert_called_once_with(
            instance_name="snapcraft-project-name", time=10
        )
        multipass_cmd().delete.assert_not_called()

    def test_mount_prime_directory(
        self, xdg_dirs, in_snap, snap_injector, multipass_cmd, base, expected_image
    ):

        with MultipassTestImpl(
            project=get_project(base), echoer=mock.Mock(), is_ephemeral=False
        ) as instance:
            instance._mount_prime_directory()

        multipass_cmd().mount.assert_any_call(
            source=mock.ANY,
            target="snapcraft-project-name:/root/prime",
            uid_map={str(os.getuid()): "0"},
            gid_map={str(os.getgid()): "0"},
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testtools.matchers import Equals

from snapcraft.internal.build_providers import errors
from snapcraft.internal.build_providers._multipass._instance_info import (  # noqa: E501
    InstanceInfo,
)
from tests import unit


class InstanceInfoGeneralTest(unit.TestCase):
    def test_initialize(self):
        instance_info = InstanceInfo(
            name="instance-name",
            state="RUNNING",
            image_release="16.04 LTS",
            mounts=dict(),
        )

        self.assertThat(instance_info.name, Equals("instance-name"))
        self.assertThat(instance_info.state, Equals("RUNNING"))
        self.assertThat(instance_info.image_release, Equals("16.04 LTS"))
        self.assertThat(instance_info.is_stopped(), Equals(False))

    def test_instance_is_stopped(self):
        instance_info = InstanceInfo(
            name="instance-name",
            state="STOPPED",
            image_release="16.04 LTS",
            mounts=dict(),
        )

        self.assertThat(instance_info.is_stopped(), Equals(True))

    def test_instance_is_mounted(self):
        instance_info = InstanceInfo(
            name="instance-name",
            state="STOPPED",
            image_release="16.04 LTS",
            mounts=dict(project=dict()),
        )

        self.assertThat(instance_info.is_mounted("project"), Equals(True))
        self.assertThat(instance_info.is_mounted("not-project"), Equals(False))


class InstanceInfoFromJSONTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.json_string = dedent(
            """\
            {
                "errors": [
                ],
                "info": {
                    "announceable-rodney": {
                        "disks": {
                            "sda1": {
                                "total": "5136297984",
                                "used": "1076072448"
                            }
                        },
                        "image_hash": "bc990872f070249450fc187d668bf65e2f87ffa3d7cf6e2934cb84b62af368e1",
                        "image_release": "16.04 LTS",
                        "ipv4": [
                            "10.74.70.141"
                        ],
                        "load": [
                            0,
                            0,
                            0
                        ],
                        "memory": {
                            "total": 1040318464,
                            "used": 42754048
                        },
                        "mounts": {
                        },
                        "release": "Ubuntu 16.04.4 LTS",
                        "state": "RUNNING"
                    }
                }
            }
        """
        )  # noqa: E501
        self.instance_name = "announceable-rodney"

    def test_new_instance_from_unmarshalled_json(self):
        instance_info = InstanceInfo.from_json(
            instance_name=self.instance_name, json_info=self.json_string
        )

        self.assertThat(instance_info.name, Equals(self.instance_name))
        self.assertThat(instance_info.state, Equals("RUNNING"))
        self.assertThat(instance_info.image_release, Equals("16.04 LTS"))

    def test_new_instance_from_unmarshalled_json_wrong_instance_fails(self):
        raised = self.assertRaises(
            errors.ProviderInfoDataKeyError,
            InstanceInfo.from_json,
            instance_name="wrong-name",
            json_info=self.json_string,
        )
        self.assertThat(raised.provider_name, Equals("multipass"))
        self.assertThat(raised.missing_key, Equals("'wrong-name'"))

    def test_new_instance_from_unmarshalled_json_empty_fails(self):
        raised = self.assertRaises(
            errors.ProviderBadDataError,
            InstanceInfo.from_json,
            instance_name=self.instance_name,
            json_info="bad-json",
        )
        self.assertThat(raised.provider_name, Equals("multipass"))
        self.assertThat(raised.data, Equals("bad-json"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Marius Gripsgard (mariogrip@ubuntu.com)
# Copyright (C) 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import os
import subprocess
from pathlib import Path
from unittest import mock

import pytest
import toml
from testtools.matchers import Contains, Equals, FileExists, Not

import snapcraft
from snapcraft.internal import errors, meta
from snapcraft.plugins.v1 import rust
from tests import fixture_setup, unit

from . import PluginsV1BaseTestCase


class RustPluginBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.useFixture(fixture_setup.CleanEnvironment())

        class Options:
            makefile = None
            make_parameters = []
            rust_features = []
            rust_revision = ""
            rust_channel = ""
            source_subdir = ""

        self.options = Options()

        patcher = mock.patch("snapcraft.internal.common.run")
        self.run_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.internal.common.run_output")
        patcher.start()
        self.addCleanup(patcher.stop)

        original_exists = os.path.exists

        def exists_mock(*args, **kwargs):
            if args[0].endswith("rustup"):
                return False
            else:
                return original_exists(args[0])

        patcher = mock.patch("os.path.exists", side_effect=exists_mock)
        patcher.start()
        self.addCleanup(patcher.stop)

        self.plugin = rust.RustPlugin("test-part", self.options, self.project)
        os.makedirs(self.plugin.sourcedir)
        os.makedirs(self.plugin.builddir)
        open(Path(self.plugin.builddir, "Cargo.toml"), "w").write("")


class RustPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        schema = rust.RustPlugin.schema()

        properties = schema["properties"]
        self.assertTrue(
            "rust-channel" in properties,
            'Expected "rust-channel" to be included in properties',
        )
        self.assertTrue(
            "rust-revision" in properties,
            'Expected "rust-revision to be included in properties',
        )

        # rust-channel
        rust_channel = properties["rust-channel"]
        self.assertTrue(
            "type" in rust_channel, 'Expected "type" to be included in "rust-channel"'
        )
        rust_channel_type = rust_channel["type"]
        self.assertThat(
            rust_channel_type,
            Equals("string"),
            'Expected "rust-channel" "type" to be "string", '
            'but it was "{}"'.format(rust_channel_type),
        )
        self.assertTrue(
            "enum" in rust_channel, 'Expected "enum" to be included in "rust-channel"'
        )
        rust_channel_enum = rust_channel["enum"]
        self.assertThat(
            rust_channel_enum,
            Equals(["stable", "beta", "nightly"]),
            'Expected "rust-channel" "enum" to be "["stable", "beta", "nightly"]", '
            'but it was "{}"'.format(rust_channel_enum),
        )

        # rust-revision
        rust_revision = properties["rust-revision"]
        self.assertTrue(
            "type" in rust_revision, 'Expected "type" to be included in "rust-revision"'
        )
        rust_revision_type = rust_revision["type"]
        self.assertThat(
            rust_revision_type,
            Equals("string"),
            'Expected "rust-revision" "type" to be "string", '
            'but it was "{}"'.format(rust_revision_type),
        )


@pytest.fixture
def options():
    class Options:
        makefile = None
        make_parameters = []
        rust_features = []
        rust_revision = ""
        rust_channel = "stable"
        source_subdir = ""

    return Options()


class TestRustPluginCrossCompile:

    scenarios = [
        ("armv7l", dict(deb_arch="armhf", target="armv7-unknown-linux-gnueabihf")),
        ("aarch64", dict(deb_arch="arm64", target="aarch64-unknown-linux-gnu")),
        ("i386", dict(deb_arch="i386", target="i686-unknown-linux-gnu")),
        ("x86_64", dict(deb_arch="amd64", target="x86_64-unknown-linux-gnu")),
        ("ppc64le", dict(deb_arch="ppc64el", target="powerpc64le-unknown-linux-gnu")),
        ("s390x", dict(deb_arch="s390x", target="s390x-unknown-linux-gnu")),
    ]

    @mock.patch("snapcraft.internal.sources._script.Script.download")
    def test_cross_compile(
        self,
        mock_download,
        monkeypatch,
        tmp_work_path,
        mock_run,
        mock_run_output,
        options,
        deb_arch,
        target,
    ):
        monkeypatch.setattr(snapcraft.project.Project, "is_cross_compiling", True)
        project = snapcraft.project.Project(target_deb_arch=deb_arch)
        project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

        plugin = rust.RustPlugin("test-part", options, project)

        with mock.patch("os.path.exists", return_value=False):
            plugin.pull()

        assert mock_run.call_count == 4
        mock_run.assert_has_calls(
            [
                mock.call(
                    [
                        os.path.join(plugin._rustup_dir, "rustup.sh"),
                        "-y",
                        "--no-modify-path",
                        "--profile=minimal",
                        "--default-toolchain",
                        "none",
                    ],
                    env=mock.ANY,
                ),
                mock.call([plugin._rustup_cmd, "install", "stable"], env=mock.ANY),
                mock.call(
                    [
                        plugin._rustup_cmd,
                        "target",
                        "add",
                        "--toolchain",
                        "stable",
                        target,
                    ],
                    env=mock.ANY,
                ),
                mock.call(
                    [
                        plugin._cargo_cmd,
                        "+stable",
                        "fetch",
                        "--manifest-path",
                        os.path.join(plugin.sourcedir, "Cargo.toml"),
                    ],
                    env=mock.ANY,
                ),
            ]
        )

        mock_run.reset_mock()
        cargo_path = Path(plugin.builddir) / "Cargo.toml"
        cargo_path.parent.mkdir(parents=True)
        cargo_path.touch()

        plugin.build()

        assert os.path.exists(os.path.join(plugin.builddir, ".cargo", "config"))
        mock_run.assert_called_once_with(
            [
                plugin._cargo_cmd,
                "+stable",
                "install",
                "--path",
                plugin.builddir,
                "--root",
                plugin.installdir,
                "--force",
                "--target",
                target,
            ],
            cwd=plugin.builddir,
            env=plugin._build_env(),
        )


class RustPluginTest(RustPluginBaseTest):
    def setUp(self):
        super().setUp()

        if self.project.deb_arch == "s390x":
            self.skipTest("architecture is not supported by rust")

    @mock.patch.object(rust.sources, "Script")
    def test_pull(self, script_mock):
        self.plugin.options.rust_revision = []
        self.plugin.options.rust_channel = []

        self.plugin.pull()

        self.assertThat(self.run_mock.call_count, Equals(3))

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    [
                        os.path.join(self.plugin._rustup_dir, "rustup.sh"),
                        "-y",
                        "--no-modify-path",
                        "--profile=minimal",
                        "--default-toolchain",
                        "none",
                    ],
                    cwd=os.path.join(self.plugin.partdir, "build"),
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [self.plugin._rustup_cmd, "install", "stable"],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [
                        self.plugin._cargo_cmd,
                        "+stable",
                        "fetch",
                        "--manifest-path",
                        os.path.join(self.plugin.sourcedir, "Cargo.toml"),
                    ],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
            ]
        )

    @mock.patch.object(rust.sources, "Script")
    def test_pull_with_cargo_lock(self, script_mock):
        self.plugin.options.rust_revision = []
        self.plugin.options.rust_channel = []
        with open(self.plugin.source_path / "Cargo.lock", "w") as f:
            f.write("")

        self.plugin.pull()

        self.assertThat(self.run_mock.call_count, Equals(3))

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    [
                        os.path.join(self.plugin._rustup_dir, "rustup.sh"),
                        "-y",
                        "--no-modify-path",
                        "--profile=minimal",
                        "--default-toolchain",
                        "none",
                    ],
                    cwd=os.path.join(self.plugin.partdir, "build"),
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [self.plugin._rustup_cmd, "install", "stable"],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [
                        self.plugin._cargo_cmd,
                        "+stable",
                        "fetch",
                        "--manifest-path",
                        os.path.join(self.plugin.sourcedir, "Cargo.toml"),
                        "--locked",
                    ],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
            ]
        )

    @mock.patch.object(rust.sources, "Script")
    def test_pull_with_rust_toolchain_file(self, script_mock):
        open(os.path.join(self.plugin.sourcedir, "rust-toolchain"), "w").close()

        self.plugin.options.rust_revision = []
        self.plugin.options.rust_channel = []

        self.plugin.pull()

        self.assertThat(self.run_mock.call_count, Equals(2))

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    [
                        os.path.join(self.plugin._rustup_dir, "rustup.sh"),
                        "-y",
                        "--no-modify-path",
                        "--profile=minimal",
                    ],
                    cwd=os.path.join(self.plugin.partdir, "build"),
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [
                        self.plugin._cargo_cmd,
                        "fetch",
                        "--manifest-path",
                        os.path.join(self.plugin.sourcedir, "Cargo.toml"),
                    ],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
            ]
        )

    @mock.patch.object(rust.sources, "Script")
    def test_pull_with_channel(self, script_mock):
        self.plugin.options.rust_revision = ""
        self.plugin.options.rust_channel = "nightly"

        self.plugin.pull()

        self.assertThat(self.run_mock.call_count, Equals(3))

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    [
                        os.path.join(self.plugin._rustup_dir, "rustup.sh"),
                        "-y",
                        "--no-modify-path",
                        "--profile=minimal",
                        "--default-toolchain",
                        "none",
                    ],
                    cwd=os.path.join(self.plugin.partdir, "build"),
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [self.plugin._rustup_cmd, "install", "nightly"],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [
                        self.plugin._cargo_cmd,
                        "+nightly",
                        "fetch",
                        "--manifest-path",
                        os.path.join(self.plugin.sourcedir, "Cargo.toml"),
                    ],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
            ]
        )

    @mock.patch.object(rust.sources, "Script")
    def test_pull_with_revision(self, script_mock):
        self.plugin.options.rust_revision = "1.13.0"
        self.plugin.options.rust_channel = ""

        self.plugin.pull()

        self.assertThat(self.run_mock.call_count, Equals(3))

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    [
                        os.path.join(self.plugin._rustup_dir, "rustup.sh"),
                        "-y",
                        "--no-modify-path",
                        "--profile=minimal",
                        "--default-toolchain",
                        "none",
                    ],
                    cwd=os.path.join(self.plugin.partdir, "build"),
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [self.plugin._rustup_cmd, "install", "1.13.0"],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [
                        self.plugin._cargo_cmd,
                        "+1.13.0",
                        "fetch",
                        "--manifest-path",
                        os.path.join(self.plugin.sourcedir, "Cargo.toml"),
                    ],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
            ]
        )

    @mock.patch.object(rust.sources, "Script")
    def test_pull_with_source_and_source_subdir(self, script_mock):
        self.plugin.options.source_subdir = "test-subdir"

        self.plugin.pull()

        self.assertThat(self.run_mock.call_count, Equals(3))

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    [
                        os.path.join(self.plugin._rustup_dir, "rustup.sh"),
                        "-y",
                        "--no-modify-path",
                        "--profile=minimal",
                        "--default-toolchain",
                        "none",
                    ],
                    cwd=os.path.join(self.plugin.partdir, "build"),
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [self.plugin._rustup_cmd, "install", "stable"],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
                mock.call(
                    [
                        self.plugin._cargo_cmd,
                        "+stable",
                        "fetch",
                        "--manifest-path",
                        os.path.join(
                            self.plugin.sourcedir, "test-subdir", "Cargo.toml"
                        ),
                    ],
                    cwd=self.plugin.builddir,
                    env=self.plugin._build_env(),
                ),
            ]
        )

    @mock.patch("snapcraft.ProjectOptions.deb_arch", "fantasy-arch")
    def test_unsupported_target_arch_raises_exception(self):
        self.assertRaises(errors.SnapcraftEnvironmentError, self.plugin._get_target)

    def test_build(self):
        self.plugin.build()

        self.run_mock.assert_called_once_with(
            [
                self.plugin._cargo_cmd,
                "+stable",
                "install",
                "--path",
                self.plugin.builddir,
                "--root",
                self.plugin.installdir,
                "--force",
            ],
            cwd=os.path.join(self.plugin.partdir, "build"),
            env=self.plugin._build_env(),
        )

    def test_install_workspace_artifacts(self):
        release_path = Path(self.plugin.builddir, "target", "release")
        os.makedirs(release_path, exist_ok=True)

        p_nonexec = Path(release_path / "nonexec")
        open(p_nonexec, "w").write("")
        p_nonexec.chmod(0o664)

        p_exec = Path(release_path / "exec")
        open(p_exec, "w").write("")
        p_exec.chmod(0o755)

        p_exec_so = Path(release_path / "exec.so")
        open(p_exec_so, "w").write("")
        p_exec_so.chmod(0o755)

        self.plugin._install_workspace_artifacts()

        bindir = Path(self.plugin.installdir, "bin")
        bins = list(bindir.iterdir())

        libdir = Path(self.plugin.installdir, "usr", "lib", self.project.arch_triplet)
        libs = list(libdir.iterdir())

        self.assertThat(bins, Equals([bindir / "exec"]))
        self.assertThat(libs, Equals([libdir / "exec.so"]))

    def test_build_workspace(self):
        cargo_path = Path(self.plugin.builddir, "Cargo.toml")
        with open(cargo_path, "w") as cargo_file:
            cargo_file.write("[workspace]" + os.linesep)
        release_path = Path(self.plugin.builddir, "target", "release")
        os.makedirs(release_path, exist_ok=True)

        self.plugin.build()

        self.run_mock.assert_called_once_with(
            [self.plugin._cargo_cmd, "+stable", "build", "--release"],
            cwd=os.path.join(self.plugin.partdir, "build"),
            env=self.plugin._build_env(),
        )

    def test_build_with_rust_toolchain_file(self):
        open(os.path.join(self.plugin.sourcedir, "rust-toolchain"), "w").close()

        self.plugin.build()

        self.run_mock.assert_called_once_with(
            [
                self.plugin._cargo_cmd,
                "install",
                "--path",
                self.plugin.builddir,
                "--root",
                self.plugin.installdir,
                "--force",
            ],
            cwd=os.path.join(self.plugin.partdir, "build"),
            env=self.plugin._build_env(),
        )

    def test_build_with_conditional_compilation(self):
        self.plugin.options.rust_features = ["conditional-compilation"]

        self.plugin.build()

        self.run_mock.assert_called_once_with(
            [
                self.plugin._cargo_cmd,
                "+stable",
                "install",
                "--path",
                self.plugin.builddir,
                "--root",
                self.plugin.installdir,
                "--force",
                "--features",
                "conditional-compilation",
            ],
            cwd=os.path.join(self.plugin.partdir, "build"),
            env=self.plugin._build_env(),
        )

    def test_build_with_cargo_lock(self):
        with open(self.plugin.source_path / "Cargo.lock", "w") as f:
            f.write("")

        self.plugin.build()

        self.run_mock.assert_called_once_with(
            [
                self.plugin._cargo_cmd,
                "+stable",
                "install",
                "--path",
                self.plugin.builddir,
                "--root",
                self.plugin.installdir,
                "--force",
                "--locked",
            ],
            cwd=os.path.join(self.plugin.partdir, "build"),
            env=self.plugin._build_env(),
        )

    def test_get_manifest_with_cargo_lock_file(self):
        with open(
            os.path.join(self.plugin.builddir, "Cargo.lock"), "w"
        ) as cargo_lock_file:
            cargo_lock_file.write("test cargo lock contents")

        self.plugin.build()

        self.assertThat(
            self.plugin.get_manifest()["cargo-lock-contents"],
            Equals("test cargo lock contents"),
        )

    def test_get_manifest_with_unexisting_cargo_lock(self):
        self.plugin.build()

        self.assertThat(
            self.plugin.get_manifest(), Not(Contains("cargo-lock-contents"))
        )

    def test_workspace(self):
        self.plugin.build()

        self.assertThat(
            self.plugin.get_manifest(), Not(Contains("cargo-lock-contents"))
        )

    def test_get_manifest_with_cargo_lock_dir(self, *_):
        os.mkdir(os.path.join(self.plugin.builddir, "Cargo.lock"))

        self.plugin.build()

        self.assertThat(
            self.plugin.get_manifest(), Not(Contains("cargo-lock-contents"))
        )

    def test_get_manifest_with_versions(self):
        original_check_output = subprocess.check_output

        def side_effect(cmd, *args, **kwargs):
            if cmd[-1] == "--version":
                binary = os.path.basename(cmd[0])
                return "test {} version".format(binary)
            return original_check_output(cmd, *args, **kwargs)

        with mock.patch.object(rust.RustPlugin, "run_output") as run_output_mock:
            run_output_mock.side_effect = side_effect
            self.plugin.build()

        expected_manifest = collections.OrderedDict()
        expected_manifest["rustup-version"] = "test rustup version"
        expected_manifest["rustc-version"] = "test rustc version"
        expected_manifest["cargo-version"] = "test cargo version"

        self.assertThat(self.plugin.get_manifest(), Equals(expected_manifest))

    def test_write_cargo_config(self):
        config_toml_path = "config.toml"

        self.plugin._write_cargo_config(cargo_config_path=config_toml_path)

        self.assertThat(config_toml_path, FileExists())

        config = toml.load(config_toml_path)

        self.assertThat(
            config,
            Equals(
                dict(
                    rustdoc_cmd=self.plugin._rustdoc_cmd,
                    rustc_cmd=self.plugin._rustc_cmd,
                    arch_triplet=self.plugin.project.arch_triplet,
                    jobs=self.plugin.parallel_build_count,
                    target={
                        self.plugin._get_target(): dict(
                            linker=self.plugin._get_linker()
                        )
                    },
                )
            ),
        )

    def test_unsupported_base(self):
        self.project._snap_meta.base = "unsupported-base"

        raised = self.assertRaises(
            errors.PluginBaseError,
            rust.RustPlugin,
            "test-part",
            self.options,
            self.project,
        )

        self.assertThat(raised.part_name, Equals("test-part"))
        self.assertThat(raised.base, Equals("unsupported-base"))



import os

from setuptools import setup

package_name = "listener_py"

setup(
    name=package_name,
    version="1.0.0",
    py_modules=["listener"],
    data_files=[
        (os.path.join("share", package_name), ["package.xml"]),
        (
            os.path.join("share", package_name, "launch"),
            [os.path.join("launch", "talk_and_listen.launch.py")],
        ),
        (
            os.path.join("share", "ament_index", "resource_index", "packages"),
            [os.path.join("resource", package_name)],
        ),
    ],
    install_requires=["setuptools"],
    entry_points={"console_scripts": ["listener = listener:main"]},
)
